/*
 * Copyright (c) 2023-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

/*
 * Abstract:
 *	Bitmap implementation.
 */

#include <string.h>
#include <complib/cl_types.h>
#include <complib/cl_byteswap_osd.h>
#include <complib/cl_bitmap.h>

#ifdef __cplusplus
#  define BEGIN_C_DECLS extern "C" {
#  define END_C_DECLS   }
#else				/* !__cplusplus */
#  define BEGIN_C_DECLS
#  define END_C_DECLS
#endif				/* __cplusplus */

BEGIN_C_DECLS

#define	_LENGTH8(X)	(((X) + 7) / 8)
#define	_LENGTH(X)	(((X) + 63) / 64)
#define	_INDEX(X)	((X) / 64)
#define	_OFFSET(X)	((X) % 64)
#define	_MASK(X)	(1ull << _OFFSET(X))

#define	_ALL_BITS	0xffffffffffffffffull


static int	_indices_set = 0;
static int	_indices_1[256][9];
static int	_indices_0[256][9];

static inline void _cl_bitmap_set_indices()
{
	int		i;
	int		index;
	int		mask;
	int		bit;
	size_t		len;

	/*
	 * Set the barrier values.
	 */
	for (index = 0; index < 256; index++) {
		_indices_1[index][0] = -1;
		_indices_0[index][0] = -1;
	}

	/*
	 * Set the '1' bit positions.
	 */
	for (index = 0; index < 256; index++) {
		i = 1;
		for (bit = 0, mask = 1; mask < 256; bit++, mask <<= 1) {
			if ((index & mask) != 0) {
				_indices_1[index][i++] = bit;
			}
		}
		_indices_1[index][0] = i-1;
	}

	/*
	 * Set the '0' bit positions.
	 */
	len = sizeof(_indices_1[0]);
	for (index = 0; index < 256; index++) {
		memcpy(_indices_0[index], _indices_1[index ^ 0xff], len);
	}

	_indices_set = 1;
}


static int	_counts[] = { 0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4, 
			      1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 
			      1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 
			      2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
			      1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 
			      2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
			      2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
			      3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 
			      1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 
			      2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
			      2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
			      3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 
			      2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
			      3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 
			      3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 
			      4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8,  };

static inline int _cl_bitmap_count(IN uint8_t value)
{
	return _counts[value];
}

/* ---------------------------------------------------------------------------------------
 * State checkers
 */
static inline cl_status_t _cl_bitmap_init_check(IN cl_bitmap_t *p_bitmap)
{
	return (p_bitmap->inited == TRUE) ? CL_SUCCESS : CL_INVALID_STATE;
}

static inline cl_status_t _cl_bitmap_bounds_check(IN cl_bitmap_t *p_bitmap,
						  IN unsigned bit_number)
{
	if (p_bitmap->inited != TRUE) {
		return CL_INVALID_STATE;
	} else if (bit_number >= p_bitmap->max_bits) {
		return CL_INVALID_PARAMETER;
	} else {
		return CL_SUCCESS;
	}
}

/* ---------------------------------------------------------------------------------------
 * Bit operations
 */
int _cl_bitmap_positions(IN cl_bitmap_t *p_bitmap,
			 IN int	p_indices[][9],
			 IN uint64_t mask,
			 OUT int *p_data,
			 IN size_t length)
{
	int		i;
	int		j;
	int		k;
	int		count;
	int		offset;
	int		*set_bits;
	uint8_t		byte;
	uint64_t	word;

	count = 0;

	for (i = 0; i < p_bitmap->len; i++) {
		offset = 64*i;
		word = p_bitmap->bits[i];
		if (word != mask) {
			for (j = 0 ;j < 8; j++) {
				byte = (uint8_t)(word & 0xff);

				set_bits = p_indices[byte];
				for (k = 1; k <= set_bits[0]; k++) {
					p_data[count++] = offset + set_bits[k];
				}

				offset += 8;
				word >>= 8;
			}
		}

	}

	return count;
}

/* ---------------------------------------------------------------------------------------
 * Logical operations
 */
typedef uint64_t (*f_lop)(IN uint64_t a, IN uint64_t b);

static inline uint64_t f_and(IN uint64_t a, IN uint64_t b)
{
	return a & b;
}

static inline uint64_t f_or(IN uint64_t a, IN uint64_t b)
{
	return a | b;
}

static inline uint64_t f_xor(IN uint64_t a, IN uint64_t b)
{
	return a ^ b;
}

static cl_status_t _cl_bitmap_lop(IN cl_bitmap_t *p_bitmap,
				  IN cl_bitmap_t *p_other,
				  IN f_lop f)
{
	int		i;

	CL_ASSERT(_cl_bitmap_bounds_check(p_bitmap, 0) == CL_SUCCESS);
	CL_ASSERT(_cl_bitmap_bounds_check(p_other, 0) == CL_SUCCESS);

	if (p_bitmap->max_bits != p_other->max_bits) {
		return CL_ERROR;
	}

	for (i = 0; i < p_bitmap->len; i++) {
		p_bitmap->bits[i] = f(p_bitmap->bits[i], p_other->bits[i]);
	}

	p_bitmap->bits[p_bitmap->len-1] &= p_bitmap->mask;
	return CL_SUCCESS;
}

static boolean_t _cl_bitmap_cop(IN cl_bitmap_t *p_bitmap, uint64_t value, boolean_t retval)
{
	int		i;
	int		extra;
	uint64_t	mask = -1ull;


	CL_ASSERT(_cl_bitmap_init_check(p_bitmap) == CL_SUCCESS);

	for (i = 0; i < p_bitmap->len-1; i++) {
		if ((p_bitmap->bits[i] & mask) != value) {
			return !retval;
		}
	}

	extra = p_bitmap->max_bits%64;
	if (extra != 0) {
		mask  = mask  >> (64 - extra);
		value = value >> (64 - extra);
	}

	if (p_bitmap->bits[p_bitmap->len-1] != value) {
		return !retval;
	}

	return retval;
}

/* =======================================================================================
 * API functions
 */

cl_status_t cl_bitmap_init(IN cl_bitmap_t *p_bitmap,
			   IN unsigned int num_bits)
{
	int		num_words;
	uint64_t	*map;

	CL_ASSERT(p_bitmap != NULL);
	CL_ASSERT((num_bits != 0) && ((num_bits%8) == 0));

	if (_indices_set == 0) {
		_cl_bitmap_set_indices();
	}

	if (p_bitmap == NULL) {
		return CL_INVALID_PARAMETER;
	}

	if ((num_bits == 0) || ((num_bits%8) != 0)) {
		return CL_INVALID_PARAMETER;
	}

	num_words = _LENGTH(num_bits);
	if ((map = calloc(1, num_words*sizeof(uint64_t))) == NULL) {
		return CL_INSUFFICIENT_MEMORY;
	}

	memset((void *)p_bitmap, 0, sizeof(cl_bitmap_t));
	p_bitmap->inited = TRUE;
	p_bitmap->max_bits = num_bits;
	p_bitmap->len = num_words;
	p_bitmap->bits = map;
	p_bitmap->mask = ((num_bits%64) == 0) ? _ALL_BITS : (1ull << (num_bits%64)) - 1;

	return CL_SUCCESS;
}

cl_status_t cl_bitmap_clone(IN cl_bitmap_t *p_dst_bitmap,
			    IN cl_bitmap_t *p_src_bitmap)
{
	cl_status_t	status;

	CL_ASSERT(_cl_bitmap_init_check(p_src_bitmap) == CL_SUCCESS);

	if (_cl_bitmap_init_check(p_dst_bitmap) == CL_SUCCESS) {
		cl_bitmap_destroy(p_dst_bitmap);
	}

	status = cl_bitmap_init(p_dst_bitmap, p_src_bitmap->max_bits);
	if (status != CL_SUCCESS) {
		return status;
	}

	memcpy(p_dst_bitmap->bits, p_src_bitmap->bits, p_src_bitmap->len * sizeof(uint64_t));
	return CL_SUCCESS;
}

void cl_bitmap_reset(IN cl_bitmap_t *p_bitmap)
{

	CL_ASSERT(_cl_bitmap_init_check(p_bitmap) == CL_SUCCESS);

	memset((void *)p_bitmap->bits, 0, p_bitmap->len * sizeof(uint64_t));
}

cl_status_t cl_bitmap_destroy(IN cl_bitmap_t *p_bitmap)
{

	CL_ASSERT(_cl_bitmap_init_check(p_bitmap) == CL_SUCCESS);

	p_bitmap->inited = FALSE;
	free(p_bitmap->bits);
	p_bitmap->bits = NULL;
	return CL_SUCCESS;
}

int cl_bitmap_get(IN cl_bitmap_t *p_bitmap,
		  IN unsigned bit_number)
{
	int		index;
	int		offset;

	CL_ASSERT(_cl_bitmap_bounds_check(p_bitmap, bit_number) == CL_SUCCESS);

	index  = _INDEX(bit_number);
	offset = _OFFSET(bit_number);
	return (p_bitmap->bits[index] >> offset) & 1;
}

cl_status_t cl_bitmap_set(IN cl_bitmap_t *p_bitmap,
			  IN unsigned bit_number)
{
	int		index;
	uint64_t	mask;

	CL_ASSERT(_cl_bitmap_bounds_check(p_bitmap, bit_number) == CL_SUCCESS);

	index = _INDEX(bit_number);
	mask  = _MASK(bit_number);

	p_bitmap->bits[index] |= mask;
	return CL_SUCCESS;
}

cl_status_t cl_bitmap_clear(IN cl_bitmap_t *p_bitmap,
			    IN unsigned bit_number)
{
	int		index;
	uint64_t	mask;

	CL_ASSERT(_cl_bitmap_bounds_check(p_bitmap, bit_number) == CL_SUCCESS);

	index = _INDEX(bit_number);
	mask  = _MASK(bit_number);

	p_bitmap->bits[index] &= ~mask;
	return CL_SUCCESS;
}

cl_status_t cl_bitmap_and(IN cl_bitmap_t *p_bitmap,
			  IN cl_bitmap_t *p_other)
{
	return _cl_bitmap_lop(p_bitmap, p_other, f_and);
}

cl_status_t cl_bitmap_or(IN cl_bitmap_t *p_bitmap,
			 IN cl_bitmap_t *p_other)
{
	return _cl_bitmap_lop(p_bitmap, p_other, f_or);
}

cl_status_t cl_bitmap_xor(IN cl_bitmap_t *p_bitmap,
			  IN cl_bitmap_t *p_other)
{
	return _cl_bitmap_lop(p_bitmap, p_other, f_xor);
}

boolean_t cl_bitmap_all(IN cl_bitmap_t *p_bitmap)
{
	return _cl_bitmap_cop(p_bitmap, -1ull, TRUE);
}

boolean_t cl_bitmap_any(IN cl_bitmap_t *p_bitmap)
{
	return _cl_bitmap_cop(p_bitmap, 0ull, FALSE);
}

int cl_bitmap_count(IN cl_bitmap_t *p_bitmap)
{
	int		i;
	int		count;
	uint8_t		*p_data;

	CL_ASSERT(_cl_bitmap_init_check(p_bitmap) == CL_SUCCESS);

	count = 0;
	p_data = (uint8_t *)p_bitmap->bits;
	for (i = 0; i < p_bitmap->len * sizeof(uint64_t); i++) {
		count += _cl_bitmap_count(*p_data);
		p_data++;
	}

	return count;
}

boolean_t cl_bitmap_is_set(IN cl_bitmap_t *p_bitmap,
			   IN unsigned bit_number)
{
	CL_ASSERT(_cl_bitmap_init_check(p_bitmap) == CL_SUCCESS);
	return cl_bitmap_get(p_bitmap, bit_number) ? TRUE : FALSE;
}

boolean_t cl_bitmap_is_clear(IN cl_bitmap_t *p_bitmap,
			     IN unsigned bit_number)
{
	CL_ASSERT(_cl_bitmap_init_check(p_bitmap) == CL_SUCCESS);
	return !cl_bitmap_get(p_bitmap, bit_number) ? TRUE : FALSE;
}

boolean_t cl_bitmap_equal(IN cl_bitmap_t *p_bitmap,
			  IN cl_bitmap_t *p_other)
{
	CL_ASSERT(_cl_bitmap_init_check(p_bitmap) == CL_SUCCESS);
	CL_ASSERT(_cl_bitmap_init_check(p_other) == CL_SUCCESS);
	CL_ASSERT(p_bitmap->max_bits == p_other->max_bits);

	return !memcmp(p_bitmap->bits, p_other->bits, p_bitmap->len * sizeof(uint64_t)) ? TRUE : FALSE;
}

cl_status_t cl_bitmap_to_array(IN cl_bitmap_t *p_bitmap,
			       OUT uint8_t *p_data,
			       IN size_t length)
{
	CL_ASSERT(_cl_bitmap_init_check(p_bitmap) == CL_SUCCESS);
	CL_ASSERT(length <= p_bitmap->max_bits);
	CL_ASSERT((length%8) == 0);

	return (memcpy(p_data, p_bitmap->bits, length/8) == p_data) ? CL_SUCCESS : CL_ERROR;
}

cl_status_t cl_bitmap_from_array(IN cl_bitmap_t *p_bitmap,
				 OUT uint8_t *p_data,
				 IN size_t length)
{
	CL_ASSERT(_cl_bitmap_init_check(p_bitmap) == CL_SUCCESS);
	CL_ASSERT(length <= p_bitmap->max_bits);
	CL_ASSERT((length%8) == 0);

	return (memcpy(p_bitmap->bits, p_data, length/8) == p_bitmap->bits) ? CL_SUCCESS : CL_ERROR;
}

int cl_bitmap_get_set_positions(IN cl_bitmap_t *p_bitmap,
				OUT int *p_data,
				IN size_t length)
{
	int		count;

	CL_ASSERT(_cl_bitmap_init_check(p_bitmap) == CL_SUCCESS);
	CL_ASSERT(length >= p_bitmap->max_bits);

	count = _cl_bitmap_positions(p_bitmap, _indices_1, 0x0ull, p_data, length);
	return count;
}

int cl_bitmap_get_cleared_positions(IN cl_bitmap_t *p_bitmap,
				    OUT int *p_data,
				    IN size_t length)
{
	int		count;

	CL_ASSERT(_cl_bitmap_init_check(p_bitmap) == CL_SUCCESS);
	CL_ASSERT(length >= p_bitmap->max_bits);

	count = _cl_bitmap_positions(p_bitmap, _indices_0, 0xffffffffffffffffull, p_data, length);
	return count;
}

END_C_DECLS
