/*
 * Copyright (c) 2024-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *	Module for planarization feature
 *
 * Author:
 *	Gull Sharon, NVIDIA
 */
#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_PLANARIZED_C
#include <opensm/osm_sm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_port.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_unhealthy_ports.h>

extern boolean_t is_sm_port(IN osm_sm_t *sm, IN osm_node_t * p_node, IN const uint8_t port_num);

boolean_t osm_aport_need_update(IN osm_aport_t * p_aport)
{
	osm_physp_t *p_physp;
	uint8_t plane;

	if (!p_aport || !osm_aport_is_valid(p_aport))
		return FALSE;

	for (plane = p_aport->min_found_plane_number; plane <= p_aport->max_found_plane_number; plane++) {
		p_physp = p_aport->plane_ports[plane];
		if (p_physp && p_physp->need_update)
			return TRUE;
	}
	return FALSE;
}

static const char *aports_system_type_strings[OSM_APORTS_SYSTEM_TYPE_MAX] = {
	"switch",
	"ca"
};

const char * osm_aports_system_get_type_str(IN const osm_aports_system_t * p_aports_system)
{
	return aports_system_type_strings[p_aports_system->type];
}

boolean_t osm_planar_link_is_internal_system_fnm_link(IN struct osm_subn * p_subn,
						      IN struct osm_physp * p_physp)
{
	osm_node_t *p_node, *p_remote_node;
	osm_physp_t *p_remote_physp;

	if (!p_physp)
		return FALSE;

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (!p_remote_physp)
		return FALSE;

	p_node = osm_physp_get_node_ptr(p_physp);
	p_remote_node = osm_physp_get_node_ptr(p_remote_physp);

	if (osm_node_get_type(p_node) != IB_NODE_TYPE_SWITCH ||
	    osm_node_get_type(p_remote_node) != IB_NODE_TYPE_SWITCH)
		return FALSE;

	if (!osm_physp_is_fnm(p_subn, p_physp) || !osm_physp_is_fnm(p_subn, p_remote_physp))
		return FALSE;

	if (p_node->node_info.sys_guid != p_remote_node->node_info.sys_guid)
		return FALSE;

	return TRUE;
}

/*
 * Given a physp of a host connected to a switch, this returns the AN physp connected to this
 * switch, and NULL if it can't be found.
 * i.e. given the following links: host -- switch -- AN,
 * an_physp = host_physp_get_neighbor_an_physp(host_physp)
 */
static osm_physp_t * host_physp_get_neighbor_an_physp(osm_physp_t *p_physp) {
	osm_physp_t *p_remote_physp, *p_sw_an_physp, *p_an_physp;

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (!p_remote_physp || !osm_physp_is_valid(p_remote_physp) ||
	    !p_remote_physp->p_node || !p_remote_physp->p_node->sw ||
	    p_remote_physp->p_node->sw->an_port_num == OSM_NO_PATH)
		return NULL;

	p_sw_an_physp = osm_node_get_physp_ptr(p_remote_physp->p_node,
					       p_remote_physp->p_node->sw->an_port_num);
	if (!p_sw_an_physp || !osm_physp_is_valid(p_sw_an_physp))
		return NULL;

	p_an_physp = osm_physp_get_remote(p_sw_an_physp);
	if (!p_an_physp || !osm_physp_is_valid(p_an_physp))
		return NULL;

	return p_an_physp;
}

static ib_api_status_t
calc_end_port_plane_filter_config(osm_sm_t *p_sm, osm_port_t *p_port, boolean_t is_prism,
				  ib_end_port_plane_filter_config_t *p_endport_plane_filter)
{
	osm_aport_t *p_aport = osm_physp_get_aport(p_port->p_physp);
	osm_physp_t *p_physp, *p_an_physp;
	uint16_t base_lid_ho, max_lid_ho, lid_ho;
	uint8_t plane, entry;

	memset(p_endport_plane_filter, 0, sizeof(ib_end_port_plane_filter_config_t));

	for (plane = p_aport->min_found_plane_number; plane <= p_aport->max_found_plane_number; plane++) {
		p_physp = p_aport->plane_ports[plane];
		if (!p_physp || !osm_physp_is_valid(p_physp)) {
			OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR B402: "
				"Port 0x%" PRIx64 " plane %u has no physp\n",
				osm_port_get_guid_ho(p_port), plane);
			return IB_ERROR;
		}

		entry = osm_physp_get_port_num(p_physp) - 1;
		OSM_ASSERT(entry < IB_END_PORT_PLANE_FILTER_CONFIG_NUM_ENTRIES);

		p_an_physp = host_physp_get_neighbor_an_physp(p_physp);
		if (!p_an_physp) {
			/*
			 * If there's no AN, the MAD should still be sent in order to clear the
			 * corresponding entry in the filter that's already configured in the host.
			 */
			OSM_LOG(p_sm->p_log, OSM_LOG_DEBUG,
				"Can't get port 0x%" PRIx64 " plane %u remote switch's AN\n",
				osm_port_get_guid_ho(p_port), plane);
			continue;
		}

		base_lid_ho = cl_ntoh16(osm_physp_get_base_lid(p_an_physp));

		if (is_prism) {
			lid_ho = base_lid_ho + plane - 1;
			max_lid_ho = base_lid_ho + (1 << osm_physp_get_lmc(p_an_physp)) - 1;

			if (lid_ho > max_lid_ho) {
				OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR B405: "
					"AN 0x%" PRIx64 " has LID range %u-%u, "
					"but LID %u is needed for plane %u\n",
					cl_ntoh64(osm_physp_get_port_guid(p_an_physp)), base_lid_ho,
					max_lid_ho, lid_ho, plane);
				return IB_ERROR;
			}
		} else {
			/* For AN on non-prism switch, just take base LID */
			lid_ho = base_lid_ho;
		}

		ib_end_port_plane_filter_config_set_lid_ho(p_endport_plane_filter, entry, lid_ho);
	} /* Planes loop */

	return IB_SUCCESS;
}

static void send_end_port_plane_filter_config(osm_sm_t *p_sm, osm_port_t *p_port,
					      ib_end_port_plane_filter_config_t *p_endport_plane_filter)
{
	osm_aport_t *p_aport = osm_physp_get_aport(p_port->p_physp);
	osm_end_port_plane_filter_config_context_t mad_context;
	ib_api_status_t status;

	if (!p_sm->p_subn->need_update && !osm_aport_need_update(p_aport) &&
	    !memcmp(p_endport_plane_filter,
		    &p_port->planarized_data.subn_end_port_plane_filter_config,
		    sizeof(ib_end_port_plane_filter_config_t)))
		return;

	/* Clear subn data, to make sure MAD is resent if current MAD has a timeout. */
	memset(&p_port->planarized_data.subn_end_port_plane_filter_config, 0,
	       sizeof(ib_end_port_plane_filter_config_t));

	memset(&mad_context, 0, sizeof(mad_context));
	mad_context.port_guid = osm_port_get_guid(p_port);

	/* Can send it to a single physp, since this will set it for the whole aport */
	status = osm_req_set(p_sm, osm_physp_get_dr_path_ptr(p_sm, p_port->p_physp),
			     p_port->p_physp, (const uint8_t *)p_endport_plane_filter,
			     sizeof(ib_end_port_plane_filter_config_t),
			     IB_MAD_ATTR_END_PORT_PLANE_FILTER_CONFIG, 0, FALSE,
			     ib_port_info_get_m_key(&p_port->p_physp->port_info), 0,
			     CL_DISP_MSGID_NONE, (osm_madw_context_t *)&mad_context);

	if (status != IB_SUCCESS)
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR B406: "
			"Setting EndPortPlaneFilterConfig of port 0x%" PRIx64 " failed (%s)\n",
			osm_port_get_guid_ho(p_port), ib_get_err_str(status));
}

void osm_planar_set_end_port_plane_filter_config(IN cl_map_item_t * p_item, IN void * context)
{
	osm_sm_t *p_sm = (osm_sm_t *)context;
	osm_port_t *p_port = (osm_port_t *)p_item;
	osm_aport_t *p_aport;
	osm_physp_t *p_remote_physp;
	ib_end_port_plane_filter_config_t endport_plane_filter;
	boolean_t is_prism;

	/* Relevant only for HCAs. */
	if (!p_port->p_node || osm_node_get_type(p_port->p_node) != IB_NODE_TYPE_CA)
		return;

	if (!ib_general_info_is_cap_enabled(&p_port->p_node->general_info,
					    IB_GENERAL_INFO_IS_END_PORT_PLANE_FILTER_CONFIG_SUPPORTED_BIT))
		return;

	/* Relevant only for planarized ports. */
	if (!p_port->p_physp || !osm_hierarchy_is_planarized_physp(p_port->p_physp))
		return;

	p_aport = osm_physp_get_aport(p_port->p_physp);
	if (!p_aport || !osm_aport_is_valid(p_aport)) {
		OSM_LOG(p_sm->p_log, OSM_LOG_ERROR, "ERR B407: "
			"Planarized port 0x%" PRIx64 " has no aport. "
			"Cannot set EndPortPlaneFilterConfig\n",
			osm_port_get_guid_ho(p_port));
		return;
	}

	if (osm_hierarchy_is_asymmetric_aport(p_aport))
		return;

	/* Relevant only if remote is a switch. */
	p_remote_physp = osm_physp_get_remote(p_port->p_physp);
	if (!p_remote_physp || !p_remote_physp->p_node || !p_remote_physp->p_node->sw)
		return;

	is_prism = osm_switch_is_prism_switch(p_remote_physp->p_node->sw, NULL);
	if (calc_end_port_plane_filter_config(p_sm, p_port, is_prism, &endport_plane_filter) !=
	    IB_SUCCESS)
		return;

	send_end_port_plane_filter_config(p_sm, p_port, &endport_plane_filter);
}

static void destroy_aports_systems_table(IN cl_qmap_t * p_aports_system_id_tbl)
{
	cl_map_item_t * p_map_item;
	osm_aports_system_t *p_aports_system;

	while (cl_qmap_count(p_aports_system_id_tbl)) {
		p_map_item = cl_qmap_head(p_aports_system_id_tbl);
		cl_qmap_remove_item(p_aports_system_id_tbl, p_map_item);
		p_aports_system = (osm_aports_system_t *)p_map_item;
		osm_aports_system_delete(&p_aports_system);
	}
}

void osm_aports_systems_tables_init(IN osm_aports_systems_tables_t * p_aports_systems_tables)
{
	cl_qmap_init(&p_aports_systems_tables->switch_aports_systems);
	cl_qmap_init(&p_aports_systems_tables->ca_aports_systems);
}

void osm_aports_systems_tables_destroy(IN osm_aports_systems_tables_t * p_aports_systems_tables)
{
	destroy_aports_systems_table(&p_aports_systems_tables->ca_aports_systems);
	destroy_aports_systems_table(&p_aports_systems_tables->switch_aports_systems);
}

osm_switch_system_t * osm_switch_get_planarized_system(IN struct osm_subn * p_subn,
						       IN struct osm_switch * p_sw)
{
	cl_qmap_t *p_sw_aports_systems_tbl;
	osm_switch_system_t *p_sw_system;
	ib_net64_t sys_guid;

	p_sw_aports_systems_tbl = &p_subn->aports_systems_tables.switch_aports_systems;
	sys_guid = p_sw->p_node->node_info.sys_guid;

	p_sw_system = (osm_switch_system_t *) cl_qmap_get(p_sw_aports_systems_tbl, sys_guid);
	if (p_sw_system == (osm_switch_system_t *)cl_qmap_end(p_sw_aports_systems_tbl))
		return NULL;

	return p_sw_system;
}

osm_physp_t* osm_planar_get_physp_by_plane(IN osm_physp_t *p_physp, IN uint8_t plane)
{
	osm_aport_t *p_aport = osm_physp_get_aport(p_physp);

	if (!p_aport || !plane || plane > IB_MAX_NUM_OF_PLANES_XDR)
		return NULL;
	return p_aport->plane_ports[plane];
}

uint8_t osm_planar_get_node_port_plane(IN osm_node_t * p_node, IN uint8_t port_num)
{
	osm_physp_t *p_physp = osm_node_get_physp_ptr(p_node, port_num);

	if (!p_physp || !osm_hierarchy_is_planarized_physp(p_physp) ||
	    p_physp->hierarchy_info.plane_number < 0)
		return 0;
	return p_physp->hierarchy_info.plane_number;
}

osm_node_t* osm_planar_get_remote_node_by_plane(IN osm_subn_t * p_subn,
						IN osm_port_t * p_port,
						IN uint8_t plane,
						OUT uint8_t * p_remote_port_num)
{
	osm_node_t *p_node = osm_port_get_node(p_port);
	uint8_t port_num;
	osm_physp_t *p_plane_physp = NULL;

	OSM_LOG_ENTER(&p_subn->p_osm->log);
	OSM_ASSERT(p_node);
	port_num = osm_physp_get_port_num(osm_port_get_physp(p_port));

	if (plane && osm_hierarchy_is_planarized_physp(p_port->p_physp)) {
		p_plane_physp = osm_planar_get_physp_by_plane(p_port->p_physp, plane);

		if (!p_plane_physp) {
			return NULL;
		}
		port_num = osm_physp_get_port_num(p_plane_physp);
	}

	OSM_LOG_EXIT(&p_subn->p_osm->log);
	return osm_node_get_remote_node(p_node, port_num, p_remote_port_num);
}

uint8_t osm_planar_get_plane_mask(IN osm_physp_t *p_physp)
{
	if (!osm_hierarchy_is_planarized_physp(p_physp))
		return 0;

	uint8_t member_num_of_planes = 0, plane_mask = 0;
	member_num_of_planes = osm_hierarchy_get_num_of_planes(p_physp);

	plane_mask = (1 << (member_num_of_planes + 1)) - 2;
	return plane_mask;
}

boolean_t osm_planar_are_all_switches_same_system(IN osm_subn_t * p_subn)
{
	osm_switch_t *p_sw;
	cl_map_item_t *sw_item;
	ib_net64_t sys_guid = 0;
	boolean_t is_single_system = TRUE;

	if (p_subn->is_planarized &&
	    cl_qmap_count(&p_subn->sw_guid_tbl) <= IB_MAX_NUM_OF_PLANES_XDR) {
		for (sw_item = cl_qmap_head(&p_subn->sw_guid_tbl);
		     sw_item != cl_qmap_end(&p_subn->sw_guid_tbl);
		     sw_item = cl_qmap_next(sw_item)) {
			p_sw = (osm_switch_t *)sw_item;
			if (!sys_guid) {
				sys_guid = p_sw->p_node->node_info.sys_guid;
				continue;
			}
			/* switches with different system guid, don't belong to the same system */
			if (p_sw->p_node->node_info.sys_guid != sys_guid) {
				is_single_system = FALSE;
			}
		}
	} else {
		is_single_system = FALSE;
	}

	return is_single_system;
}

boolean_t osm_planar_is_link_asymmetric(IN osm_subn_t * p_subn,
					IN osm_physp_t * p_physp,
					OUT boolean_t * p_is_port_asymmetric)
{
	osm_physp_t *p_physp_remote;
	boolean_t is_asymmetric_link = FALSE;

	OSM_ASSERT(p_physp);

	if (p_is_port_asymmetric)
		*p_is_port_asymmetric = FALSE;

	if (!p_subn->is_planarized)
		goto Exit;

	if (osm_hierarchy_is_asymmetric_aport_physp(p_physp)) {
		if (p_is_port_asymmetric)
			*p_is_port_asymmetric = TRUE;

		is_asymmetric_link = TRUE;
		goto Exit;
	}

	p_physp_remote = p_physp->p_remote_physp;
	if (p_physp_remote && osm_physp_is_valid(p_physp_remote) &&
	    osm_hierarchy_is_asymmetric_aport_physp(p_physp_remote))
		is_asymmetric_link = TRUE;

Exit:
	return is_asymmetric_link;
}

void osm_planar_disable_ar_to_switches_on_same_system(IN osm_subn_t * p_subn)
{
	cl_qmap_t *p_sw_aports_systems_tbl;
	cl_map_item_t *p_item;
	osm_switch_system_t *p_sw_system;
	uint8_t asic, lid_idx, port, state;
	osm_node_t *p_node;
	osm_switch_t *p_sw;
	osm_physp_t *p_physp0;
	boolean_t is_prism;
	boolean_t ar_configured_on_system;
	uint16_t *lids_on_system;
	uint16_t sw_lid, sw_base_lid, sw_max_lid, lid, group_id;
	uint16_t num_lids, system_total_num_lids;

	num_lids = (1 << osm_subn_get_max_lmc(p_subn)) * OSM_PLANAR_MAX_NUM_OF_ASICS;

	/* dense array of all lids assigned to switches on the system */
	lids_on_system = (uint16_t *)malloc(num_lids * sizeof(uint16_t));

	if (!lids_on_system) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_ERROR, "ERR B401"
			"Failed to allocate memory for switch systems lids processing.\n");
		return;
	}

	p_sw_aports_systems_tbl = &p_subn->aports_systems_tables.switch_aports_systems;

	for (p_item = cl_qmap_head(p_sw_aports_systems_tbl);
	     p_item != cl_qmap_end(p_sw_aports_systems_tbl);
	     p_item = cl_qmap_next(p_item)) {
		p_sw_system = (osm_switch_system_t *)p_item;
		memset(lids_on_system, 0, num_lids * sizeof(uint16_t));
		ar_configured_on_system = FALSE;
		/* assume switch system is prism switch system, till proven otherwise */
		is_prism = TRUE;
		system_total_num_lids = 0;

		/* get all the lids of all the switches of the planarized non-prism system */
		for (asic = p_sw_system->min_asic_available;
		     asic <= p_sw_system->max_asic_available;
		     asic ++) {
			if (!p_sw_system->nodes[asic] || !p_sw_system->nodes[asic]->sw) {
				continue;
			}

			p_node = p_sw_system->nodes[asic];
			p_sw = p_node->sw;

			is_prism = osm_switch_is_prism_switch(p_sw, NULL);

			/* skip prism switch system */
			if (is_prism) {
				break;
			}

			/* at least one of switches has ar configured */
			if (p_sw->ar_configured == OSM_SW_AR_CONF_CONFIGURED)
				ar_configured_on_system = TRUE;

			p_physp0 = osm_node_get_physp_ptr(p_node, 0);
			sw_base_lid = osm_node_get_base_lid_ho(p_node, 0);
			sw_max_lid = sw_base_lid + (1 << osm_physp_get_lmc(p_physp0)) - 1;

			for (sw_lid = sw_base_lid; sw_lid <= sw_max_lid; sw_lid++) {
				lids_on_system[system_total_num_lids] = sw_lid;
				system_total_num_lids ++;
			}
		}

		/* skip prism system or system without AR configuration */
		if (is_prism || !ar_configured_on_system) {
			continue;
		}

		for (asic = p_sw_system->min_asic_available;
		     asic <= p_sw_system->max_asic_available;
		     asic ++) {
			if (!p_sw_system->nodes[asic] || !p_sw_system->nodes[asic]->sw) {
				continue;
			}

			p_node = p_sw_system->nodes[asic];
			p_sw = p_node->sw;
			p_physp0 = osm_node_get_physp_ptr(p_node, 0);

			/* skip changing AR configuration on switch without AR configuration */
			if (p_sw->ar_configured != OSM_SW_AR_CONF_CONFIGURED)
				continue;

			sw_base_lid = osm_node_get_base_lid_ho(p_node, 0);
			sw_max_lid = sw_base_lid + (1 << osm_physp_get_lmc(p_physp0)) - 1;

			for (lid_idx = 0; lid_idx <= system_total_num_lids; lid_idx++) {

				lid = lids_on_system[lid_idx];

				/* skip non-existent switches, or lid of the current switch */
				if (!lid || (lid >= sw_base_lid && lid <= sw_max_lid))
					continue;

				/* currently support only for PLFT 0 */

				/* get static port configured for this lid */
				osm_ar_lft_entry_get(p_sw, 0, lid, &port, &state, &group_id);
				/*
				 * Set routing towards destination lid on the same system
				 * as static and without AR group (group 0).
				 */
				osm_ar_lft_set(&p_sw->ar_data, lid, 0, port,
					       OSM_AR_IB_LID_STATE_STATIC, 0);
			}
		}
	}
	free(lids_on_system);
}

static void osm_unhealthy_aport_process(IN osm_sm_t *sm,
					IN osm_aport_t *p_aport,
					IN osm_unhealthy_port_t *p_plane_uhp,
					IN uint8_t plane_mask)
{
	uint8_t plane_port;
	osm_physp_t *p_physp;
	osm_node_t *p_node;
	osm_unhealthy_port_t *p_uhp_by_aport, *p_uhp_exist;
	cl_fmap_item_t *p_item;

	for (plane_port = 1; plane_port <= p_aport->max_found_plane_number; plane_port++) {
		p_physp = p_aport->plane_ports[plane_port];

		/* skip the originally found unhealthy ports */
		if (!p_physp || (plane_mask & (1 << plane_port)))
			continue;

		p_node = p_physp->p_node;

		p_uhp_by_aport =
			osm_unhealthy_port_new(0, 0, NULL, 0,
					osm_node_get_node_guid(p_node),
					p_physp->port_num,
					p_node->print_desc);

		if (!p_uhp_by_aport) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR B408: "
				"Couldn't create new unhealthy port UNKNOWN[0]<-->0x%016"
				PRIx64 "[%d]\n",
				osm_node_get_node_guid(p_node),
				p_physp->port_num);
			return;
		}

		p_uhp_by_aport->cond_mask = p_plane_uhp->cond_mask;
		p_uhp_by_aport->report = p_plane_uhp->report;
		p_uhp_by_aport->ignore = p_plane_uhp->ignore;
		p_uhp_by_aport->disable = p_plane_uhp->disable;
		p_uhp_by_aport->isolate = p_plane_uhp->isolate;
		p_uhp_by_aport->no_discover = p_plane_uhp->no_discover;
		p_uhp_by_aport->manual_action = p_plane_uhp->manual_action;

		p_item = cl_fmap_insert(&sm->p_subn->unhealthy_ports_tbl,
				&p_uhp_by_aport->p_obj,
				&p_uhp_by_aport->fmap_item);

		uint8_t aport_num = p_physp->hierarchy_info.aport;

		if (p_item != &p_uhp_by_aport->fmap_item) {
			p_uhp_exist = (osm_unhealthy_port_t *)p_item;
			osm_unhealthy_port_remove(sm, &p_uhp_by_aport);
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR B409: "
				"APort %u on System GUID 0x%" PRIx64
				" node GUID 0x%" PRIx64	" port number %u "
				"failed to add unhealthy, existing 0x%" PRIx64
				" port %u peer 0x%" PRIx64 " port %u.\n",
				aport_num, cl_ntoh64(p_physp->p_node->node_info.sys_guid),
				osm_node_get_node_guid_ho(p_physp->p_node),
				p_physp->port_num, cl_ntoh64(p_uhp_exist->guid), p_uhp_exist->port_num,
				cl_ntoh64(p_uhp_exist->peer_guid), p_uhp_exist->peer_port_num);
		} else {
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
				"APort %u on System GUID 0x%" PRIx64
				" node GUID 0x%" PRIx64	" port number %u "
				"succeeded to add unhealthy\n",
				aport_num, cl_ntoh64(p_physp->p_node->node_info.sys_guid),
				osm_node_get_node_guid_ho(p_physp->p_node),
				p_physp->port_num);
		}
	}
}

/*
 * This function returns the previous plane symmetry state of given APort,
 * based on the previous state of physical ports (plane ports) that are associated with the APort.
 * Currently there is no way to have previous APort state in the very rare
 * corner case for planarized HCA:
 * where there are two disjoint sets of plane ports (A and B) that comprise the HCA's APort,
 * in one sweep ports from set A are down and ports from set B are up, the next sweep
 * plane ports from set A are up, while plane ports for set B are up.
 */
static osm_physp_plane_symmetric_state_t
planar_get_aport_previous_symmetry_status(IN osm_aport_t *p_aport)
{
	osm_physp_plane_symmetric_state_t state = OSM_PHYSP_PLANE_SYMMETRIC;
	uint8_t plane, port_num;
	osm_physp_t *p_physp;
	osm_node_t *p_node;

	OSM_ASSERT(p_aport && osm_aport_is_valid(p_aport));

	for (plane = p_aport->min_found_plane_number;
	     plane <= p_aport->max_found_plane_number;
	     plane++) {
		p_physp = p_aport->plane_ports[plane];
		/* skip missing plane port */
		if (!p_physp || !osm_physp_is_valid(p_physp))
			continue;

		p_node = osm_physp_get_node_ptr(p_physp);
		port_num = osm_physp_get_port_num(p_physp);

		/*
		 * at least one of the plane ports previous state is asymmetric,
		 * then the whole APort previous state is asymmetric.
		 */
		if (p_node->physp_plane_symmetric_state[port_num] == OSM_PHYSP_PLANE_ASYMMETRIC) {
			state = OSM_PHYSP_PLANE_ASYMMETRIC;
			break;
		}
	}

	return state;
}

static void
planar_set_aport_symmetry_status(IN OUT osm_subn_t *p_subn,
				 IN osm_aport_t *p_aport,
				 IN osm_physp_plane_symmetric_state_t prev_symmetry_state,
				 IN osm_physp_plane_symmetric_state_t symmetry_state)
{
	uint8_t plane, port_num;
	osm_physp_t *p_physp;
	osm_node_t *p_node;
	osm_aports_system_t *p_aports_system;

	p_aport->asymmetric = (symmetry_state == OSM_PHYSP_PLANE_ASYMMETRIC) ? TRUE : FALSE;

	for (plane = p_aport->min_found_plane_number;
	     plane <= p_aport->max_found_plane_number;
	     plane++) {
		p_physp = p_aport->plane_ports[plane];
		/* skip missing plane port */
		if (!p_physp || !osm_physp_is_valid(p_physp))
			continue;

		p_node = osm_physp_get_node_ptr(p_physp);
		port_num = osm_physp_get_port_num(p_physp);
		p_node->physp_plane_symmetric_state[port_num] = symmetry_state;
	}

	p_aports_system = osm_aport_get_aports_system(p_aport);

	if (prev_symmetry_state != symmetry_state) {
		OSM_LOG(&p_subn->p_osm->log, OSM_LOG_INFO,
			"The APort %u on system ID 0x%" PRIx64
			" changed state from %s to %s, invalidating ucast cache.\n",
			p_aport->aport_num,
			cl_ntoh64(p_aports_system->guid_id),
			prev_symmetry_state == OSM_PHYSP_PLANE_ASYMMETRIC ?
			"asymmetric" : "symmetric",
			symmetry_state == OSM_PHYSP_PLANE_ASYMMETRIC ?
			"asymmetric" : "symmetric");
		p_subn->p_osm->sm.ucast_mgr.variables_invalidate_cache = TRUE;
	}
}

static void osm_analyze_aports_system(cl_map_item_t * const p_item, void *context)
{
	osm_aports_system_t *p_aports_system = (osm_aports_system_t *)p_item;
	osm_sm_t *sm = (osm_sm_t *)context;
	osm_aport_iter_t aport_iter;
	osm_aport_t *p_aport;
	osm_physp_t *p_physp;
	osm_node_t *p_node, *p_sm_node;
	uint8_t first_aport_num_of_planes, number_of_found_planes, unhealthy_planes_mask;
	uint8_t first_plane_port, plane_port, first_port, sm_port_number, first_unhealthy_plane_port;
	uint8_t first_port_state = 0, phys_port_state = 0;
	uint16_t first_port_link_speed_active, port_link_speed_active;
	uint8_t first_link_width_active = 0, first_mtu_cap = 0, first_vl_cap = 0;
	boolean_t first_port_found = FALSE;
	osm_unhealthy_port_t *p_uhp, *p_first_uhp;
	boolean_t is_partially_unhealthy;
	boolean_t has_fdr10 = (sm->p_subn->opt.mepi_enabled_speeds & FDR10_BIT) ? TRUE : FALSE;
	uint64_t first_node_guid;
	boolean_t is_sm_aport, aport_asymmetric;
	osm_physp_plane_symmetric_state_t aport_prev_plane_symmetry_state;

	OSM_LOG_ENTER(sm->p_log);

	for (aport_iter = osm_aports_system_begin(p_aports_system);
	     aport_iter != osm_aports_system_end(p_aports_system);
	     aport_iter = osm_aport_iter_next(aport_iter)) {
		p_aport = (osm_aport_t *)aport_iter;

		if (!osm_aport_is_valid(p_aport))
			continue;

		/* reset for new aport */
		first_port_state = phys_port_state = first_port_link_speed_active = port_link_speed_active =
		first_link_width_active = first_mtu_cap = first_vl_cap = first_aport_num_of_planes =
		number_of_found_planes = first_node_guid = first_plane_port = first_port = sm_port_number =
			first_unhealthy_plane_port = unhealthy_planes_mask = 0;
		aport_prev_plane_symmetry_state =
			planar_get_aport_previous_symmetry_status(p_aport);

		first_port_found = FALSE;
		p_sm_node = NULL;
		is_sm_aport = FALSE;
		p_first_uhp = NULL;
		is_partially_unhealthy = FALSE;
		aport_asymmetric = FALSE;

		for (plane_port = 1; plane_port <= p_aport->max_found_plane_number; plane_port++) {
			p_physp = p_aport->plane_ports[plane_port];

			if (!p_physp || !p_physp->p_node) {
				continue;
			}

			p_node = p_physp->p_node;

			if (!is_sm_aport && is_sm_port(sm, p_node, p_physp->port_num)) {
				is_sm_aport = TRUE;
				p_sm_node = p_node;
				sm_port_number = p_physp->port_num;
			}

			/*
			 * look at parameters:
			 * Relevant attribute values are:
			 * active speed, active width, MTU cap and VL cap.
			 */
			phys_port_state = ib_port_info_get_port_state(&p_physp->port_info);
			number_of_found_planes++;

			p_uhp = osm_hm_get_unhealthy_port_by_guid(sm,
								  osm_node_get_node_guid(p_node),
								  p_physp->port_num);

			if (p_uhp && (p_uhp->isolate || p_uhp->no_discover)) {
				/*
				 * if all plane ports of this APort are unhealthy, then everything
				 * is symmetrical.
				 */
				if (!p_first_uhp) {
					/*
					 * This section is applied only for the first unhealthy plane port
					 * found for this APort
					 */
					p_first_uhp = p_uhp;
					first_unhealthy_plane_port = plane_port;

					/* other healthy planes exist on this APort */
					if (number_of_found_planes > 1)
						is_partially_unhealthy = TRUE;
				}

				/* remember the plane of unhealthy plane port */
				unhealthy_planes_mask |= (1 << plane_port);

				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Found unhealthy plane %u on APort %u on System GUID 0x%" PRIx64
					" node GUID 0x%" PRIx64	" port number %u, "
					"isolate %u no_discover %u (APort's partially unhealthy %u)\n",
					plane_port,
					p_aport->aport_num, cl_ntoh64(p_node->node_info.sys_guid),
					osm_node_get_node_guid_ho(p_node),
					p_physp->port_num,
					p_uhp->isolate, p_uhp->no_discover, is_partially_unhealthy);

			} else if (!p_uhp && p_first_uhp) {
				/* found APort that some plane ports are healthy and some are not */
				is_partially_unhealthy = TRUE;

				OSM_LOG(sm->p_log, OSM_LOG_DEBUG,
					"Found partially unhealthy APort %u on System GUID 0x%" PRIx64
					", healthy node GUID 0x%" PRIx64 " port number %u plane %u, "
					"while unhealthy for plane %u: isolate %u no_discover %u.\n",
					p_aport->aport_num, cl_ntoh64(p_node->node_info.sys_guid),
					osm_node_get_node_guid_ho(p_node),
					p_physp->port_num, plane_port, first_unhealthy_plane_port,
					p_first_uhp->isolate, p_first_uhp->no_discover);
			}

			/* first found port, lets remember its values for comparison */
			if (!first_port_found) {
				first_plane_port = plane_port;
				first_port = p_physp->port_num;
				first_port_state = phys_port_state;
				first_port_link_speed_active =
					osm_physp_get_highest_active_speed(p_physp, has_fdr10);
				first_link_width_active = p_physp->port_info.link_width_active;
				first_mtu_cap = ib_port_info_get_mtu_cap(&p_physp->port_info);
				first_vl_cap = ib_port_info_get_vl_cap(&p_physp->port_info);
				first_port_found = TRUE;
				first_aport_num_of_planes = p_physp->hierarchy_info.num_of_planes;
				first_node_guid = osm_node_get_node_guid_ho(p_node);
				continue;
			}

			port_link_speed_active =
				osm_physp_get_highest_active_speed(p_physp, has_fdr10);

			/*
			 * Check if one of the plane ports has DOWN state while others don't.
			 * Check if the APort is not marked as asymmetric yet.
			 */
			if (!aport_asymmetric &&
			    phys_port_state != first_port_state &&
			    (phys_port_state == IB_LINK_DOWN || first_port_state == IB_LINK_DOWN)) {
				aport_asymmetric = TRUE;

				/* print only when APort becomes asymmetric first time */
				if (aport_prev_plane_symmetry_state == OSM_PHYSP_PLANE_SYMMETRIC) {
					OSM_LOG(sm->p_log, OSM_LOG_INFO,
						"Found state asymmetric APort %u on System GUID 0x%016" PRIx64
						" node GUID 0x%016" PRIx64 " port number %u plane %u: "
						"state %u compared to node GUID 0x%016" PRIx64
						" port number %u plane %u: state %u.\n",
						p_aport->aport_num, cl_ntoh64(p_node->node_info.sys_guid),
						osm_node_get_node_guid_ho(p_node),
						p_physp->port_num, plane_port, phys_port_state,
						first_node_guid, first_port, first_plane_port,
						first_port_state);
				}
			}

			/* Check if the APort is not marked as asymmetric yet. */
			if (!aport_asymmetric &&
			    (port_link_speed_active != first_port_link_speed_active ||
			    p_physp->port_info.link_width_active != first_link_width_active ||
			    ib_port_info_get_mtu_cap(&p_physp->port_info) != first_mtu_cap ||
			    ib_port_info_get_vl_cap(&p_physp->port_info) != first_vl_cap ||
			    p_physp->hierarchy_info.num_of_planes != first_aport_num_of_planes)) {

				aport_asymmetric = TRUE;

				OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR B40A: "
					"Found attributes asymmetric APort %u on System GUID 0x%" PRIx64
					" node GUID 0x%" PRIx64
					" port number %u plane %u: speed %u, width %u, mtu_cap %u, vl_cap %u,"
					" number of planes %u compared to node GUID 0x%016" PRIx64
					" port number %u plane %u: speed %u, width %u, mtu_cap %u, vl_cap %u,"
					" number of planes %u.\n",
					p_aport->aport_num, cl_ntoh64(p_node->node_info.sys_guid),
					osm_node_get_node_guid_ho(p_node),
					p_physp->port_num, plane_port,
					port_link_speed_active, p_physp->port_info.link_width_active,
					ib_port_info_get_mtu_cap(&p_physp->port_info),
					ib_port_info_get_vl_cap(&p_physp->port_info),
					p_physp->hierarchy_info.num_of_planes,
					first_node_guid, first_port, first_plane_port,
					first_port_link_speed_active, first_link_width_active,
					first_mtu_cap, first_vl_cap, first_aport_num_of_planes);
			}

		} /* end of planes loop */

		/* make sure to mark all other plane ports of this APort as unhealthy */
		if (is_partially_unhealthy) {
			osm_unhealthy_aport_process(sm, p_aport,
						    p_first_uhp, unhealthy_planes_mask);
		}

		/*
		 * Check APort that wasn't marked yet as asymmetric,
		 * that all plane ports exist for it.
		 */
		if (!aport_asymmetric &&
		    (!number_of_found_planes || number_of_found_planes != first_aport_num_of_planes)) {
			aport_asymmetric = TRUE;

			/* print only when APort becomes asymmetric first time */
			if (aport_prev_plane_symmetry_state == OSM_PHYSP_PLANE_SYMMETRIC) {
				OSM_LOG(sm->p_log, OSM_LOG_INFO,
					"Found asymmetric APort %u on %s aports system ID 0x%016" PRIx64
					" with missing planes. Found %u planes out of %u planes.\n",
					p_aport->aport_num, osm_aports_system_get_type_str(p_aports_system),
					cl_ntoh64(p_aports_system->guid_id),
					number_of_found_planes, first_aport_num_of_planes);
			}
		}

		if (aport_asymmetric && is_sm_aport) {
			OSM_LOG(sm->p_log, OSM_LOG_ERROR, "ERR B40B: SM port/peer "
				"0x%" PRIx64 "[%u] found to belong to asymmetric APort %u on %s "
				"aports system 0x%" PRIx64 ". Ignoring this state (Not allowed).\n",
				osm_node_get_node_guid_ho(p_sm_node), sm_port_number,
				p_aport->aport_num, osm_aports_system_get_type_str(p_aports_system),
				cl_ntoh64(p_aports_system->guid_id));
			aport_asymmetric = FALSE;
		}

		/* update the current state of plane symmetry of each plane port of that APort */
		planar_set_aport_symmetry_status(sm->p_subn,
						 p_aport,
						 aport_prev_plane_symmetry_state,
						 aport_asymmetric ?
						 OSM_PHYSP_PLANE_ASYMMETRIC :
						 OSM_PHYSP_PLANE_SYMMETRIC);

	} /* end of APorts loop */

	OSM_LOG_EXIT(sm->p_log);
}

void osm_planar_analyze_aports(IN osm_sm_t * sm)
{
	cl_qmap_apply_func(&sm->p_subn->aports_systems_tables.switch_aports_systems,
			   osm_analyze_aports_system, sm);
	cl_qmap_apply_func(&sm->p_subn->aports_systems_tables.ca_aports_systems,
			   osm_analyze_aports_system, sm);
}

void osm_planar_decrease_system_fnm_links_number(IN OUT osm_sm_t * sm, IN osm_node_t *p_node)
{
	osm_switch_system_t *p_sw_system;

	if (osm_node_get_type(p_node) != IB_NODE_TYPE_SWITCH || !p_node->sw) {
		return;
	}

	p_sw_system = osm_switch_get_planarized_system(sm->p_subn, p_node->sw);
	if (!p_sw_system)
		return;

	if (p_sw_system->num_internal_fnm_ports > 0) {
		p_sw_system->num_internal_fnm_ports--;
	}
}

boolean_t osm_planar_is_same_non_prism_system(IN osm_subn_t *p_subn,
					      IN osm_switch_t *p_sw,
					      IN ib_net64_t guid)
{
	boolean_t is_same_system = FALSE;
	osm_switch_t *p_other_sw;
	boolean_t is_prism_sw, is_prism_other_sw;
	uint8_t sw_plane, other_sw_plane;

	if (!p_subn->is_planarized || !p_sw || !p_sw->p_node)
		goto Exit;

	is_prism_sw = osm_switch_is_prism_switch(p_sw, &sw_plane);

	/* prism or legacy switch */
	if (is_prism_sw || !sw_plane)
		goto Exit;

	p_other_sw = osm_get_switch_by_guid(p_subn, guid);

	/* no switch with a given guid exists */
	if (!p_other_sw || !p_other_sw->p_node)
		goto Exit;

	is_prism_other_sw = osm_switch_is_prism_switch(p_other_sw, &other_sw_plane);

	/* prism or legacy switch */
	if (is_prism_other_sw || !other_sw_plane)
		goto Exit;

	/* both switches are planarized non-prism switch, a.k.a Black Mamba */
	if (p_sw->p_node->node_info.sys_guid == p_other_sw->p_node->node_info.sys_guid) {
		is_same_system = TRUE;
	}

Exit:
	return is_same_system;
}

void osm_planar_set_fnm_port(osm_subn_t *p_subn, ib_net64_t guid_id, int asic, uint8_t fnm_port)
{
	if (!guid_id || asic < 1 || asic > OSM_PLANAR_MAX_NUM_OF_ASICS)
		return;

	uint8_t i;
	osm_aports_system_t *p_aports_system;
	cl_qmap_t *p_aports_system_tbl = &p_subn->aports_systems_tables.switch_aports_systems;
	osm_switch_system_t *p_sw_system;

	p_aports_system = (osm_aports_system_t *) cl_qmap_get(p_aports_system_tbl, guid_id);
	if (p_aports_system == (osm_aports_system_t *)cl_qmap_end(p_aports_system_tbl))
		return;

	p_sw_system = (osm_switch_system_t *)p_aports_system;

	/* set at the first free slot */
	for (i = 0; i < OSM_PLANAR_NUM_INTERNAL_FNM_PORTS; i++) {
		if (p_sw_system->node_fnm_ports[asic][i] == 0) {
			p_sw_system->node_fnm_ports[asic][i] = fnm_port;
			break;
		}
	}
}

boolean_t osm_planar_is_fnm_ring_broken(IN osm_switch_system_t *p_sw_system)
{
	uint8_t num_asics, num_internal_fnm_links;

	/* Each internal FNM link has two internal FNM ports. */
	num_internal_fnm_links = p_sw_system->num_internal_fnm_ports / 2;
	num_asics = p_sw_system->max_asic_available - p_sw_system->min_asic_available + 1;

	/* a complete FNM ring requires #links >= #nodes, otherwise fnm ring is broken */
	return (num_internal_fnm_links < num_asics);
}

static void
planar_set_routing_to_sw_special_ports(osm_sm_t *p_sm,
				       osm_switch_t *p_sw,
				       osm_switch_t *p_dest_sw,
				       uint8_t static_port,
				       uint8_t state,
				       uint16_t ar_group_id)
{
	uint8_t num_ports, port_num, remote_port;
	osm_node_t *p_dest_node, *p_remote_node;
	osm_physp_t *p_physp, *p_remote_physp;
	ib_special_port_type_t special_type;

	p_dest_node = p_dest_sw->p_node;

	/* add static routing to special Nodes connected directly to the switch */
	num_ports = osm_node_get_num_physp(p_dest_node);
	for (port_num = 1 ; port_num < num_ports; port_num++) {
		p_physp = osm_node_get_physp_ptr(p_dest_node, port_num);
		if (!p_physp || !osm_physp_is_valid(p_physp))
			continue;
		p_remote_node = osm_node_get_remote_node(p_dest_node, port_num, &remote_port);
		if (!p_remote_node || osm_node_get_type(p_remote_node) == IB_NODE_TYPE_SWITCH)
			continue;
		p_remote_physp = osm_physp_get_remote(p_physp);

		/* skip regular ports */
		special_type = osm_physp_get_special_port_type(p_remote_physp);
		if (special_type == IB_SPECIAL_PORT_TYPE_RESERVED)
			continue;

		if (lookup_isolated_port(p_sm, p_dest_node, port_num))
			continue;

		if (osm_planar_is_link_asymmetric(p_sm->p_subn, p_physp, NULL))
			continue;

		osm_switch_set_routing(p_sw, p_remote_node, remote_port,
				       static_port, state, ar_group_id);
	}
}

/*
 * This function adds routing from planarized non-prism (a.k.a Black Mamba) switch p_sw to
 * destination planarized non-prism switch p_dest_sw of different plane through given
 * static port and adaptive routing group.
 */
static void planar_add_routing_other_plane_switch(IN osm_sm_t *p_sm,
						  IN OUT osm_switch_t *p_sw,
						  IN osm_switch_t *p_dest_sw,
						  IN uint8_t static_port,
						  IN uint8_t state,
						  IN uint16_t ar_group_id)
{
	osm_switch_set_routing(p_sw, p_dest_sw->p_node, 0, static_port,
			       state, ar_group_id);
	/*
	 * set the same routing as to destination switch
	 * to non-planarized special HCAs (Aggregation Node, Router), that are connected to it.
	 */
	planar_set_routing_to_sw_special_ports(p_sm, p_sw, p_dest_sw, static_port, state, ar_group_id);
}

static void planar_add_fnm_hops(IN cl_map_item_t * p_map_item, IN void *context)
{
	osm_sm_t *sm = (osm_sm_t *) context;
	osm_switch_system_t *p_sw_system = (osm_switch_system_t *) p_map_item;
	uint8_t asic, i, remote_port, all_asics_set_mask;
	osm_node_t *p_node, *p_remote_node;
	osm_switch_t *p_sw;
	osm_physp_t *p_remote_physp;
	uint8_t fnm_port;
	uint16_t sw_lid, asic_to_lid[OSM_PLANAR_MAX_NUM_OF_ASICS + 1] = {0};
	uint8_t asic_to_invalid_fnm[OSM_PLANAR_MAX_NUM_OF_ASICS + 1][OSM_PLANAR_NUM_INTERNAL_FNM_PORTS] = {0};
	uint8_t asic_to_other_asic_set_mask[OSM_PLANAR_MAX_NUM_OF_ASICS + 1] = {0};
	uint8_t num_internal_fnm_links = 0, num_asics, remote_asic, least_hops, miss_asic;
	boolean_t is_isolated_port, is_no_discover_port, was_update;

	all_asics_set_mask = (1 << OSM_PLANAR_MAX_NUM_OF_ASICS) - 1;

	/* fnm link is not used for routing on prism switch, don't add hops */
	if (osm_planar_is_prism_sw_system(p_sw_system))
		return;

	/*
	 * Go over system's asics and mark what fnm links can be used.
	 * The fnm link can be used if the asic switch is not marked as held-back and
	 * if the link is not marked as 'no discover' or 'isolated'.
	 */
	for (asic = p_sw_system->min_asic_available;
	     asic <= p_sw_system->max_asic_available; asic++) {
		if (!p_sw_system->nodes[asic] || !p_sw_system->nodes[asic]->sw) {
			continue;
		}

		p_node = p_sw_system->nodes[asic];
		p_sw = p_node->sw;
		sw_lid = osm_node_get_base_lid_ho(p_sw_system->nodes[asic], 0);
		if (!sw_lid) {
			/* TODO ERR? */
			continue;
		}

		asic_to_lid[asic] = sw_lid;
		/* set min hops to itself */
		osm_switch_set_hops(p_sw, sw_lid, 0, 0);
		asic_to_other_asic_set_mask[asic] |= (1 << asic);

		if (lookup_held_back_switch(sm->p_subn, p_node)) {
			for (i = 0; i < OSM_PLANAR_NUM_INTERNAL_FNM_PORTS; i++) {
				asic_to_invalid_fnm[asic][i] = 1;
			}
			/* continue to next asic */
			continue;
		}

		for (i = 0; i < OSM_PLANAR_NUM_INTERNAL_FNM_PORTS; i++) {
			fnm_port = p_sw_system->node_fnm_ports[asic][i];
			if (!fnm_port)
				continue;
			is_isolated_port = lookup_isolated_port(sm, p_node, fnm_port);
			is_no_discover_port = lookup_no_discover_port(sm, p_node, fnm_port);
			if (is_isolated_port || is_no_discover_port) {
				asic_to_invalid_fnm[asic][i] = 1;
				continue;
			}
			p_remote_node = osm_node_get_remote_node(p_node, fnm_port, &remote_port);
			if (!p_remote_node ||
			    lookup_held_back_switch(sm->p_subn, p_remote_node)) {
				asic_to_invalid_fnm[asic][i] = 1;
				continue;
			}
			p_remote_physp = osm_node_get_physp_ptr(p_remote_node, remote_port);
			if (!p_remote_physp || !osm_physp_is_valid(p_remote_physp) ||
			    p_remote_physp->hierarchy_info.asic < 0) {
				asic_to_invalid_fnm[asic][i] = 1;
				continue;
			}
			num_internal_fnm_links ++;
		}
	}

	num_internal_fnm_links /= OSM_PLANAR_NUM_INTERNAL_FNM_PORTS;
	num_asics = p_sw_system->max_asic_available - p_sw_system->min_asic_available + 1;

	/* TODO: consider removing this fnm ring break since this code is currently called after osm_remove_links */
	/* the ring is complete, ignore one link (first asic to last asic) */
	if (num_internal_fnm_links == num_asics) {
		asic_to_invalid_fnm[p_sw_system->min_asic_available][1] = 1;
	}

	do {
		was_update = FALSE;
		for (asic = p_sw_system->min_asic_available;
		     asic <= p_sw_system->max_asic_available; asic++) {
			if (!p_sw_system->nodes[asic] || !p_sw_system->nodes[asic]->sw) {
				continue;
			}

			/* The switch has already hops to all other asics, nothing to do. */
			if (asic_to_other_asic_set_mask[asic] == all_asics_set_mask)
				continue;

			p_node = p_sw_system->nodes[asic];
			p_sw = p_node->sw;

			/* check neighbors through valid internal fnm ports */
			for (i = 0; i < OSM_PLANAR_NUM_INTERNAL_FNM_PORTS; i++) {

				fnm_port = p_sw_system->node_fnm_ports[asic][i];
				if (asic_to_invalid_fnm[asic][i] || !fnm_port)
					continue;

				p_remote_node =
					osm_node_get_remote_node(p_node, fnm_port, &remote_port);
				p_remote_physp =
					osm_node_get_physp_ptr(p_remote_node, remote_port);

				/* shouldn't happen, as such port was marked as invalid */
				if (PF(p_remote_physp->hierarchy_info.asic < 0))
					continue;

				remote_asic = p_remote_physp->hierarchy_info.asic;
				/*
				 * Already update that remote asic in the past.
				 * Check if switch can get hops to missing asics
				 * from remote neighbor asic.
				 */
				if (asic_to_other_asic_set_mask[asic] & (1 << remote_asic)) {
					/* go over missing asics */
					for (miss_asic = p_sw_system->min_asic_available;
					     miss_asic <= p_sw_system->max_asic_available;
					     miss_asic++) {
						/* switch doesn't have hops to miss_asic yet */
						if (!(asic_to_other_asic_set_mask[asic] &
							(1 << miss_asic))) {
							/* check if remote asic has hops */
							least_hops = osm_switch_get_least_hops(
									p_remote_node->sw,
									asic_to_lid[miss_asic]);

							if (least_hops == OSM_NO_PATH)
								continue;
							/* hops to miss_asic via neighbor asic */
							osm_switch_set_hops(p_sw,
									    asic_to_lid[miss_asic],
									    fnm_port,
									    (least_hops + 1));
							asic_to_other_asic_set_mask[asic] |=
								(1 << miss_asic);
							was_update = TRUE;
						}
					}
					continue;
				}
				osm_switch_set_hops(p_sw, asic_to_lid[remote_asic], fnm_port, 1);
				asic_to_other_asic_set_mask[asic] |= (1 << remote_asic);
				was_update = TRUE;
			}
		} /* end of asics loop */
	} while (was_update);
}

/*
 * This function adds routing on asics of planarized non-prism switch system (Black Mamba)
 * to other asics on the same system through FNM links.
 */
static void planar_add_fnm_routing(IN cl_map_item_t * p_map_item, IN void *context)
{
	osm_sm_t *sm = (osm_sm_t *) context;
	osm_switch_system_t *p_sw_system = (osm_switch_system_t *) p_map_item;
	uint8_t asic, other_asic;
	osm_node_t *p_node, *p_other_node;
	osm_switch_t *p_sw, *p_other_sw;
	uint16_t other_sw_lid_ho;
	uint8_t least_hops, hops;
	uint8_t fnm_port, fnm_port_to_route;
	uint8_t i;
	uint8_t min_hops;

	/* nothing routes through FNM ports on prism switches */
	p_node = p_sw_system->nodes[p_sw_system->min_asic_available];
	if (!p_node || !p_node->sw || osm_switch_is_prism_switch(p_node->sw, NULL))
		return;

	for (asic = p_sw_system->min_asic_available;
	     asic <= p_sw_system->max_asic_available; asic++) {
		if (!p_sw_system->nodes[asic] || !p_sw_system->nodes[asic]->sw) {
			continue;
		}

		p_node = p_sw_system->nodes[asic];
		p_sw = p_node->sw;

		for (other_asic = p_sw_system->min_asic_available;
		     other_asic <= p_sw_system->max_asic_available; other_asic++) {
			if (!p_sw_system->nodes[other_asic] || !p_sw_system->nodes[other_asic]->sw ||
			    p_sw_system->nodes[other_asic]->sw == p_sw) {
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "For sw %s skip other asic %u\n",
					p_sw->p_node->print_desc, other_asic);
				continue;
			}

			p_other_node = p_sw_system->nodes[other_asic];
			p_other_sw = p_other_node->sw;
			other_sw_lid_ho = osm_node_get_base_lid_ho(p_other_node, 0);

			/* add routing to other switch through internal FNM port of the ring */
			least_hops = osm_switch_get_least_hops(p_sw, other_sw_lid_ho);
			if (least_hops == OSM_NO_PATH) {
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "No route from sw %s to %s.\n",
					p_sw->p_node->print_desc, p_other_sw->p_node->print_desc);
				continue;
			}

			fnm_port_to_route = 0;
			min_hops = OSM_NO_PATH;
			for (i = 0; i < OSM_PLANAR_NUM_INTERNAL_FNM_PORTS; i ++) {
				fnm_port = p_sw_system->node_fnm_ports[asic][i];
				/* not set for some reason */
				if (!fnm_port)
					continue;
				hops = osm_switch_get_hop_count(p_sw, other_sw_lid_ho, fnm_port);
				/*
				 * There is supposed to be valid hops only through single fnm port.
				 * If the fnm ring is valid, then one of the links is removed
				 * from routing, to prevent loops.
				 */
				if (hops != OSM_NO_PATH && hops < min_hops) {
					fnm_port_to_route = fnm_port;
					min_hops = hops;
				}
			}

			if (!fnm_port_to_route) {
				OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "No route from sw %s to %s "
					"through FNM port.\n",
					p_sw->p_node->print_desc, p_other_sw->p_node->print_desc);
				continue;
			}

			planar_add_routing_other_plane_switch(sm, p_sw, p_other_sw,
							      fnm_port_to_route,
							      OSM_AR_IB_LID_STATE_STATIC, 0);
			OSM_LOG(sm->p_log, OSM_LOG_DEBUG, "Set on sw %s route to %s via port %u\n",
				p_sw->p_node->print_desc, p_other_sw->p_node->print_desc,
				fnm_port_to_route);
		}
	}
}

void osm_planar_add_fnm_routing(osm_sm_t * sm)
{
	cl_qmap_apply_func(&sm->p_subn->aports_systems_tables.switch_aports_systems,
			   planar_add_fnm_hops,
			   sm);
	cl_qmap_apply_func(&sm->p_subn->aports_systems_tables.switch_aports_systems,
			   planar_add_fnm_routing,
			   sm);
}

boolean_t osm_planar_is_prism_sw_system(osm_switch_system_t *p_sw_system)
{
	osm_node_t *p_sw_node;

	if (!p_sw_system ||
	    !p_sw_system->nodes[p_sw_system->min_asic_available] ||
	    !p_sw_system->nodes[p_sw_system->min_asic_available]->sw)
		return FALSE;

	p_sw_node = p_sw_system->nodes[p_sw_system->min_asic_available];

	return osm_switch_is_prism_switch(p_sw_node->sw, NULL);
}

/* configure switch p_sw's hops to destination switch through representative switch node p_rep_node */
static void planar_add_hops_other_plane_switch(IN osm_sm_t *p_sm,
					       IN OUT osm_switch_t *p_sw,
					       IN osm_node_t *p_rep_node,
					       IN osm_node_t *p_dest_sw_node)
{
	uint8_t port_num;
	uint16_t rep_sw_node_lid, dest_sw_lid;
	uint8_t least_hops_rep_to_dest, least_hops, hops_to_rep;
	uint8_t num_hops = OSM_NO_PATH;

	if (!p_sw || !p_rep_node || !p_rep_node->sw || !p_dest_sw_node || !p_dest_sw_node->sw)
		return;

	rep_sw_node_lid = osm_node_get_base_lid_ho(p_rep_node, 0);
	dest_sw_lid = osm_node_get_base_lid_ho(p_dest_sw_node, 0);

	least_hops_rep_to_dest = osm_switch_get_least_hops(p_rep_node->sw, dest_sw_lid);
	least_hops = osm_switch_get_least_hops(p_sw, rep_sw_node_lid);

	/* representative has no hops to destination or switch has no hops to representative node */
	if (least_hops_rep_to_dest == OSM_NO_PATH || least_hops == OSM_NO_PATH)
		return;

	/*
	 * Complete hops from switch to destination through all ports that have hops
	 * to representative node.
	 */
	for (port_num = 1; port_num < p_sw->num_ports; port_num++) {
		hops_to_rep = osm_switch_get_hop_count(p_sw, rep_sw_node_lid, port_num);

		if (hops_to_rep == OSM_NO_PATH)
			continue;

		/* hops from sw to representative and then from it to destination */
		num_hops = hops_to_rep + least_hops_rep_to_dest;
		osm_switch_set_hops(p_sw, dest_sw_lid, port_num, num_hops);
	}
}

/*
 * Set routing between different non-prism (AKA Black Mamba) systems.
 * Add missing routing between asics of different planes.
 * Complete route from system A asic x to another system B asic y (x and y are different planes)
 * through route from x to B's asic on plane x (representative asic on system B),
 * only if there is route from B's x to B's y inside B via FNMs.
 */
static void
planar_route_non_prism_to_non_prism(osm_sm_t *p_sm,
				    osm_switch_system_t *p_sw_system,
				    osm_switch_system_t *p_other_sw_system)
{
	uint8_t node_idx, other_node_idx, sw_plane, plane;
	osm_node_t *p_sw_node, *p_other_sw_node, *p_other_rep_sw_node;
	osm_node_t* plane_to_other_system_node[IB_MAX_NUM_OF_PLANES_XDR + 1] = {NULL};
	uint16_t other_sw_lid;
	osm_switch_t *p_sw;
	uint8_t plft_id = 0; /* Currently only plft 0 */
	uint8_t rep_port, rep_state;
	uint8_t port, state;
	uint16_t rep_group_id, group_id;

	if (p_sw_system == p_other_sw_system)
		return;

	for (other_node_idx = p_other_sw_system->min_asic_available;
	     other_node_idx <= p_other_sw_system->max_asic_available;
	     other_node_idx++) {
		p_other_sw_node = p_other_sw_system->nodes[other_node_idx];
		if (!p_other_sw_node || !p_other_sw_node->sw)
			continue;

		osm_switch_is_prism_switch(p_other_sw_node->sw, &sw_plane);

		plane_to_other_system_node[sw_plane] = p_other_sw_node;
	}

	for (node_idx = p_sw_system->min_asic_available;
	     node_idx <= p_sw_system->max_asic_available;
	     node_idx ++) {
		p_sw_node = p_sw_system->nodes[node_idx];
		if (!p_sw_node || !p_sw_node->sw)
			continue;

		p_sw = p_sw_node->sw;

		osm_switch_is_prism_switch(p_sw, &sw_plane);
		/* TODO ERR? */

		if (!sw_plane || !plane_to_other_system_node[sw_plane])
			continue;

		p_other_rep_sw_node = plane_to_other_system_node[sw_plane];
		other_sw_lid = osm_node_get_base_lid_ho(p_other_rep_sw_node, 0);

		/*
		 * Get routing from switch to switch on the same plane on another system
		 * (representative node A on another system).
		 */
		osm_ar_lft_entry_get(p_sw, plft_id, other_sw_lid, &rep_port, &rep_state, &rep_group_id);
		/* if there is no routing to representative, skip completing routes through it */
		if (rep_port == OSM_NO_PATH) {
			continue;
		}

		/* copy the same routing to other system's planes */
		for (plane = 1; plane <= IB_MAX_NUM_OF_PLANES_XDR; plane++) {
			/* skip routing to plane that we got routing from */
			if (plane == sw_plane || !plane_to_other_system_node[plane] ||
			    !plane_to_other_system_node[plane]->sw)
				continue;

			p_other_sw_node = plane_to_other_system_node[plane];
			other_sw_lid = osm_node_get_base_lid_ho(p_other_sw_node, 0);

			/*
			 * Complete route towards other plane asic B on another system through
			 * the representative asic A, only if representative A has route to B.
			 */
			osm_ar_lft_entry_get(p_other_rep_sw_node->sw, plft_id, other_sw_lid,
					     &port, &state, &group_id);
			if (port == OSM_NO_PATH)
				continue;

			planar_add_routing_other_plane_switch(p_sm, p_sw, p_other_sw_node->sw,
							      rep_port, rep_state, rep_group_id);

			/* complete hops through representative node switch */
			planar_add_hops_other_plane_switch(p_sm, p_sw, p_other_rep_sw_node, p_other_sw_node);
		}
	}
}

static void planar_add_systems_routing(cl_map_item_t * const p_item, void *context)
{
	osm_switch_system_t *p_sw_system = (osm_switch_system_t *)p_item;
	osm_switch_system_t *p_other_sw_system;
	osm_sm_t *sm = (osm_sm_t *)context;
	cl_qmap_t *p_sw_aports_systems_tbl;
	cl_map_item_t *p_other_item;
	boolean_t is_prism_sw;

	is_prism_sw = osm_planar_is_prism_sw_system(p_sw_system);

	/* if it's prism switch - nothing to do */
	if (is_prism_sw)
		return;

	p_sw_aports_systems_tbl = &sm->p_subn->aports_systems_tables.switch_aports_systems;

	for (p_other_item = cl_qmap_head(p_sw_aports_systems_tbl);
	     p_other_item != cl_qmap_end(p_sw_aports_systems_tbl);
	     p_other_item = cl_qmap_next(p_other_item)) {
		p_other_sw_system = (osm_switch_system_t *)p_other_item;

		if (osm_planar_is_prism_sw_system(p_other_sw_system))
			continue;

		planar_route_non_prism_to_non_prism(sm, p_sw_system, p_other_sw_system);
	}

}

void osm_planar_add_systems_routing(IN osm_sm_t * sm)
{
	cl_qmap_apply_func(&sm->p_subn->aports_systems_tables.switch_aports_systems,
			   planar_add_systems_routing, sm);
}

/*
 * Check whether this port can pass unicast traffic received on the specified plane.
 * Prism switch can receive traffic in all 4 planes, but it can also have 2-planes hosts or
 * legacy switches/hosts connected to it.
 * In such case traffic can flow through the prism switch from the 4 planes into 2 planes or
 * into non-planarized ports, according to the following logic:
 * - Ports can pass traffic received on plane X if they also belong to plane X.
 * - Non-planarized ports can pass traffic received on all planes.
 * - Ports belonging to 2-planes APort that belong to plane X can pass traffic received on
 *   plane X and on plane X+2. I.e. 1 <-> 1 or 3 and 2 <-> 2 or 4.
 * This logic is derived from the way FW set up the entry plane filter, which maps the plane
 * the traffic entered through into the planes it can exit through. The entry plane filter is
 * configured by FW, and its logic is mimicked here.
 * Note that this only affects unicast routing. Multicast routing uses different logic.
 */
boolean_t osm_planar_port_routable_in_plane(IN osm_node_t * p_node,
					    IN uint32_t port_num,
					    IN uint8_t plane)
{
	osm_physp_t *p_physp = osm_node_get_physp_ptr(p_node, port_num);
	int port_plane, num_of_planes, plane_for_comparison;

	if (!p_physp || !osm_physp_is_valid(p_physp))
		return FALSE;

	/* Non-planarized port can pass traffic coming from any plane */
	if (!osm_hierarchy_is_planarized_physp(p_physp))
		return TRUE;

	port_plane = p_physp->hierarchy_info.plane_number;
	num_of_planes = p_physp->hierarchy_info.num_of_planes;

	/*
	 * If port has 2 planes, it can still receive traffic coming from planes 3-4.
	 * Plane 1 receives traffic from planes 1 and 3,
	 * while plane 2 receives traffic from planes 2 and 4.
	 */
	plane_for_comparison = plane;
	if (num_of_planes == IB_MIN_NUM_OF_PLANES_XDR &&
	    (IB_MIN_NUM_OF_PLANES_XDR < plane && plane <= IB_MAX_NUM_OF_PLANES_XDR))
		plane_for_comparison = plane - IB_MIN_NUM_OF_PLANES_XDR;

	return port_plane == plane_for_comparison;
}

boolean_t osm_planar_port_connectivity_correct(IN osm_node_t * p_node, IN uint32_t port_num)
{
	osm_physp_t *p_physp, *p_remote_physp;
	osm_aport_t *p_aport, *p_remote_aport;
	uint8_t plane;
	boolean_t both_unplanarized, any_unplanarized, both_missing, any_missing;

	p_physp = osm_node_get_physp_ptr(p_node, port_num);
	if (!p_physp || !osm_physp_is_valid(p_physp))
		return FALSE;

	p_remote_physp = osm_physp_get_remote(p_physp);
	if (!p_remote_physp || !osm_physp_is_valid(p_remote_physp))
		return FALSE;

	both_unplanarized = (!osm_hierarchy_is_planarized_physp(p_physp) &&
			     !osm_hierarchy_is_planarized_physp(p_remote_physp));
	any_unplanarized = (!osm_hierarchy_is_planarized_physp(p_physp) ||
			    !osm_hierarchy_is_planarized_physp(p_remote_physp));

	/* If one side is not planarized, the other side should also be unplanarized. */
	if (any_unplanarized) {
		return both_unplanarized;
	}

	p_aport = osm_physp_get_aport(p_physp);
	p_remote_aport = osm_physp_get_aport(p_remote_physp);
	if (!p_aport || !osm_aport_is_valid(p_aport) ||
	    !p_remote_aport || !osm_aport_is_valid(p_remote_aport))
		return FALSE;

	if (p_physp->hierarchy_info.plane_number != p_remote_physp->hierarchy_info.plane_number)
		return FALSE;

	/* Required for looping over planes */
	if ((p_aport->max_found_plane_number != p_remote_aport->max_found_plane_number) ||
	    (p_aport->min_found_plane_number != p_remote_aport->min_found_plane_number))
		return FALSE;

	/* Check all planes are correctly connected */
	for (plane = p_aport->min_found_plane_number; plane <= p_aport->max_found_plane_number; plane++) {
		p_physp = p_aport->plane_ports[plane];
		p_remote_physp = p_remote_aport->plane_ports[plane];

		both_missing = (!p_physp && !p_remote_physp);
		any_missing = (!p_physp || !p_remote_physp);

		/* Missing plane port. */
		if (any_missing) {
			/* Both plane ports are missing. */
			if (both_missing)
				continue;

			/* Only one side is missing. */
			return FALSE;
		}

		/* Same plane, but the ports are not connected to each other. */
		if (osm_physp_get_remote(p_physp) != p_remote_physp)
			return FALSE;
	}

	return TRUE;
}
