/*
 * Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Main entry point into version 2 routing subsystem.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <math.h>
#include <regex.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/time.h>

#include <complib/cl_types_osd.h>
#include <complib/cl_types.h>
#include <complib/cl_math.h>
#include <complib/cl_byteswap.h>
#include <complib/cl_heap.h>
#include <complib/cl_bitmap.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTING_V2_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_routing.h>
#include <opensm/osm_advanced_routing.h>
#include <opensm/osm_graph.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_routing_v2.h>
#include <opensm/osm_helper.h>

/* -------------------------------------------------------------------------------------------------
 * Routing engine private definitions.
 */
typedef int(*osm_routing_function_t)(osm_route_context_t *);
typedef int(*osm_routing_sw_function_t)(osm_route_context_t *, osm_vertex_t *);
typedef void *(*osm_routing_sw_init_function_t)(osm_route_context_t *);
typedef void *(*osm_routing_sw_fini_function_t)(void *);

/* ----------------------------------------------------------------------------------------------
 * Debug section.
 */

#define v2_snprintf(format, ...)	bufused += snprintf(buffer+bufused, bufsize-bufused, format, ## __VA_ARGS__)

#define print_or_log(BUFFER) {									\
	if (route_debug) {									\
		printf("%s", BUFFER);								\
	}											\
												\
	if (p_log != NULL) {									\
		OSM_LOG(p_log, OSM_LOG_DEBUG, "%s", BUFFER);					\
	}											\
}

static boolean_t		route_debug = FALSE;

__attribute__((unused))
static void route_print(IN osm_route_context_t *p_context,
			IN osm_graph_t *p_graph,
			IN osm_vertex_t *p_src_vertex,
			IN osm_vertex_t *p_dst_vertex)
{
	int			i;
	int			count;
	int			ports[IB_ENDP_MAX_NUMBER_PORTS];
	size_t			bufsize;
	size_t			bufused;
	char			buffer[4096];
	osm_log_t		*p_log;
	osm_switch_t		*p_src_sw;
	osm_switch_t		*p_dst_sw;
	osm_route_t		*p_route;

	p_log = osm_log_is_active_v2(p_graph->p_log, OSM_LOG_DEBUG, FILE_ID) ? p_graph->p_log : NULL;
	if (!route_debug && (p_log == NULL)) {
		return;
	}

	bufused = 0;
	bufsize = sizeof(buffer);

	p_route = osm_vertex_to_vertex_route(p_context, p_src_vertex, p_dst_vertex);
	p_src_sw = osm_graph_vertex_to_switch(p_src_vertex);
	p_dst_sw = osm_graph_vertex_to_switch(p_dst_vertex);
	v2_snprintf("\t%-8s -> %-8s : ", osm_switch_name(p_src_sw), osm_switch_name(p_dst_sw));
	if ((int32_t)p_route->heap_index >= 0) {
		v2_snprintf("H=%3d  : ", (int32_t)p_route->heap_index);
	} else {
		v2_snprintf("       : ");
	}

	count = cl_bitmap_get_set_positions(&p_route->ports, ports, IB_ENDP_MAX_NUMBER_PORTS);
	for (i = 0; i < count; i++) {
		v2_snprintf("%2d  ", ports[i]);
	}

	v2_snprintf("\n");
	print_or_log(buffer);
}

__attribute__((unused))
static void route_print_all(osm_route_context_t *p_context)
{
	size_t			bufsize;
	size_t			bufused;
	char			buffer[256];
	osm_log_t		*p_log;
	osm_vertex_t		*p_src_vertex;
	osm_vertex_t		*p_dst_vertex;
	osm_graph_t		*p_graph;

	p_graph = p_context->p_graph;

	p_log = osm_log_is_active_v2(p_graph->p_log, OSM_LOG_DEBUG, FILE_ID) ? p_graph->p_log : NULL;

	if (!route_debug && (p_log == NULL)) {
		return;
	}

	bufused = 0;
	bufsize = sizeof(buffer);

	print_or_log("============================================================================================\n");
	v2_snprintf("GRAPH[%d] routes\n", p_graph->plane);
	print_or_log(buffer);

	foreach_graph_vertex(p_graph, p_src_vertex) {
		foreach_graph_vertex(p_graph, p_dst_vertex) {
			route_print(p_context, p_graph, p_src_vertex, p_dst_vertex);
		}
		print_or_log("\n");
	}
	print_or_log("\n");
}

/* ----------------------------------------------------------------------------------------------
 * Utility section.
 */

typedef enum {
	MINHOP_PORT_COUNT	= 0,
	MINHOP_PORT_NUMS	= 1,
	MINHOP_PORT_MASK	= 2,
} minhop_info_type_t;

static uint8_t routing_get_minhop_info(IN osm_route_context_t *p_context,
				       IN osm_vertex_t *p_src_vertex,
				       IN osm_vertex_t *p_dst_vertex,
				       IN void *p_void,
				       IN minhop_info_type_t type)
{
	int			index;
	int			offset;
	uint8_t			port_count;
	uint8_t			least_hops;
	uint8_t			edge_least_hops;
	uint8_t			*port_list;
	uint64_t		*port_bitmap;
	osm_edge_t		*p_edge;

	if (type == MINHOP_PORT_NUMS) {
		port_list = (uint8_t *)p_void;
		memset(port_list, 0, IB_ENDP_MAX_NUMBER_PORTS);
	} else if (type == MINHOP_PORT_MASK) {
		port_bitmap = (uint64_t *)p_void;
		memset(port_bitmap, 0, ROUTING_PORT_MASK_BYTES);
	}

	least_hops = osm_routing_get_least_hops(p_context, p_src_vertex, p_dst_vertex);
	if (least_hops == OSM_NO_PATH) {
		return 0;
	}

	port_count = 0;

	foreach_graph_vertex_edge(p_src_vertex, p_edge) {
		edge_least_hops = osm_routing_get_hops(p_context, p_src_vertex, p_dst_vertex, p_edge);
		if (edge_least_hops == least_hops) {
			if (type == MINHOP_PORT_NUMS) {
				port_list[port_count] = p_edge->port_num;
			} else if (type == MINHOP_PORT_MASK) {
				index = p_edge->port_num / 64;
				offset = p_edge->port_num % 64;

				port_bitmap[index] |= 1ull << offset;
			}

			port_count++;

		}
	}

	return port_count;
}

/* ---------------------------------------------------------------------------------------------- */

static int route_vertices(IN void **tasks,
			  IN unsigned task_len,
			  IN void *p_context,
			  IN unsigned int thread_id)
{
	int			i;
	int			status = 0;
	osm_vertex_t		*p_vertex;
	osm_ucast_mgr_t		*p_mgr;
	osm_route_context_t	context;

	context = *(osm_route_context_t *)p_context;
	p_mgr = context.p_mgr;

	OSM_LOG_ENTER(p_mgr->p_log);

	if (context.sw_init_function) {
		context.task_context = ((osm_routing_sw_init_function_t)context.sw_init_function)(&context);
		if (context.task_context == NULL) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR BD01: allocation failed for visited and routes\n");
			return 1;
		}
	}

	for (i = 0; i < task_len; i++) {
		p_vertex = (osm_vertex_t *)tasks[i];
		((osm_routing_sw_function_t)context.sw_function)(&context, p_vertex);
	}

	if (context.sw_fini_function) {
		((osm_routing_sw_fini_function_t)context.sw_fini_function)(context.task_context);
	}

	OSM_LOG_EXIT(p_mgr->p_log);
	return status;
}

/* ----------------------------------------------------------------------------------------------
 * Heap section.
 */

void osm_routing_index_update(const void * context, const size_t new_index)
{
	osm_route_t		*p_route = (osm_route_t *)context;

	if (p_route) {
		p_route->heap_index = new_index;
	}
}

int osm_routing_key_cmp(const void *p_lhs, const void *p_rhs)
{
	uint64_t		key1 = *(uint64_t *)p_lhs;
	uint64_t		key2 = *(uint64_t *)p_rhs;

	return (key1 < key2) ? -1 : 1;
}

/* ----------------------------------------------------------------------------------------------
 * API section.
 */

int osm_routing_execute_threaded(IN osm_route_context_t *p_context,
				 IN void *function,
				 IN void *p_func_context)
{
	osm_vertex_t		*p_vertex;
	osm_graph_t		*p_graph;
	osm_ucast_mgr_t		*p_mgr;
        void                    **tasks;

	p_mgr = p_context->p_mgr;
	p_graph = p_context->p_graph;

	/*
	 * Allocate space for task pointers.
	 */
	if ((tasks = (void **)calloc(p_graph->num_vertices, sizeof(void *))) == NULL) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR BD03: allocation failed for tasks\n");
		return -1;
	}

	/*
	 * Put the vertices into the task list.
	 */
	foreach_graph_vertex(p_graph, p_vertex) {
		tasks[p_vertex->index] = (void *)p_vertex;
	}

	/*
	 * Execute the function.
	 */
	cl_mp_thread_pool_array_apply(&p_mgr->thread_pool,
			tasks,
			p_graph->num_vertices,
			CL_MP_THREAD_POOL_BATCH_PER_THREAD,
			(cl_pfn_mp_thread_pool_apply_t)function,
			p_func_context);

	/*
	 * Cleanup.
	 */
	free(tasks);
	return 0;
}

int osm_routing_per_vertex(osm_route_context_t *p_context)
{
	return osm_routing_execute_threaded(p_context, route_vertices, p_context);
}

int osm_routing_get_ports(IN osm_route_context_t *p_context,
			  IN osm_switch_t *p_src_sw,
			  IN osm_switch_t *p_dst_sw,
			  OUT uint64_t *ports)
{
	int			count = 0;
	osm_vertex_t		*p_src_vertex;
	osm_vertex_t		*p_dst_vertex;
	osm_route_t		*p_route;

	p_src_vertex = osm_graph_switch_to_vertex(p_context->p_graph, p_src_sw);
	p_dst_vertex = osm_graph_switch_to_vertex(p_context->p_graph, p_dst_sw);

	memset(ports, 0, ROUTING_PORT_MASK_BYTES);

	/*
	 * If there is no route, return 0.
	 */
	p_route = osm_vertex_to_vertex_route(p_context, p_src_vertex, p_dst_vertex);
	if (p_route == NULL) {
		return 0;
	}

	/*
	 * Get the ports.
	 */
	osm_route_ports_to_array(p_route, (uint8_t *)ports, IB_ENDP_MAX_NUMBER_PORTS);
	count = osm_route_ports_count(p_route);

	/*
	 * In the case where the count is zero, there are two cases:
	 *	1) no route exists - like UPDN with core -> core
	 *	2) routing engine didn't set the route
	 *
	 * In either case, we will try to get the minhop ports so that a route can be used.
	 * If the count is still zero, then the ucast manager will have to stitch together
	 * two routes (through the hub switch) to make the connection.
	 */
	if (count == 0) {
		count = osm_routing_get_minhop_ports(p_context, p_src_vertex, p_dst_vertex, ports);
	}

	return count;
}

int osm_routing_v2(IN osm_route_context_t *p_context, IN osm_routing_engine_t *p_re)
{
	osm_vertex_t		*p_src_vertex;
	osm_vertex_t		*p_dst_vertex;
	osm_graph_t		*p_graph;
	osm_route_t		*p_route;
	int			status;

	OSM_LOG_ENTER(p_context->p_mgr->p_log);

	OSM_LOG(p_context->p_mgr->p_log, OSM_LOG_DEBUG, "Routing plane %u\n", p_context->plane);

	/*
	 * Setup the routes.
	 */
	p_context->p_re = p_re;
	p_graph = p_context->p_graph;

	foreach_graph_vertex(p_graph, p_src_vertex) {
		foreach_graph_vertex(p_graph, p_dst_vertex) {
			p_route = osm_vertex_to_vertex_route(p_context, p_src_vertex, p_dst_vertex);
			if (p_route != NULL) {
				p_route->src = p_src_vertex->index;
				p_route->dst = p_dst_vertex->index;
				p_route->heap_index = UINT32_MAX;
			}
		}
	}

	/*
	 * Run the algorithm.
	 */
	status = p_re->route(p_context);
	if (status != 0) {
		goto Exit;
	}

	route_print_all(p_context);

Exit:
	OSM_LOG_EXIT(p_context->p_mgr->p_log);
	return status;
}

void osm_routing_v2_destroy(IN osm_ucast_mgr_t *p_mgr, IN int plane)
{
	size_t			iter;
	osm_route_t		*p_route;
	osm_route_context_t	*p_context;

	OSM_LOG_ENTER(p_mgr->p_log);

	p_context = p_mgr->v2_context[plane];;

	if (p_context != NULL) {
		foreach_route(p_context, p_route, iter) {
			osm_route_ports_destroy(p_route);
		}

		cl_bitmap_destroy(&p_context->least_dirs);

		free(p_context->p_hops);
		free(p_context->p_least_hops);
		free(p_context->p_routes);

		free(p_context);
	}

	p_mgr->v2_context[plane] = NULL;

	OSM_LOG_EXIT(p_mgr->p_log);
}

osm_route_context_t *osm_routing_v2_create(IN osm_ucast_mgr_t *p_mgr, IN int plane)
{
	size_t			iter;
	size_t			num_square;
	size_t			num_vertices;
	uint8_t			*p_least_hops;
	osm_hops_t		*p_hops;
	osm_graph_t		*p_graph;
	osm_route_t		*p_route;
	osm_route_t		*p_routes;
	osm_route_context_t	*p_context;
	osm_route_context_t	*ret;

	OSM_LOG_ENTER(p_mgr->p_log);

	p_graph = osm_ucast_mgr_plane_to_graph(p_mgr, plane);
	num_vertices = p_graph->num_vertices;
	num_square = num_vertices * num_vertices;

	p_hops = (osm_hops_t *)calloc(num_vertices, sizeof(osm_hops_t));
	p_least_hops = (uint8_t *)calloc(num_square, sizeof(uint8_t));
	p_routes = (osm_route_t *)calloc(num_vertices * num_vertices, sizeof(osm_route_t));
	p_context = (osm_route_context_t *)calloc(1, sizeof(osm_route_context_t));

	if (osm_any_null(4, p_hops, p_least_hops, p_routes, p_context)) {
		free(p_hops);
		free(p_least_hops);
		free(p_routes);
		free(p_context);
		ret = NULL;
		goto Exit;
	}

	p_context->p_mgr = p_mgr;
	p_context->p_graph = p_graph;
	p_context->plane = plane;
	p_context->num_vertices = num_vertices;
	p_context->p_routes = p_routes;
	p_context->p_hops = p_hops;
	p_context->p_least_hops = p_least_hops;

	cl_bitmap_init(&p_context->least_dirs, num_vertices * ROUNDUP(num_vertices, 64));

	foreach_route(p_context, p_route, iter) {
		osm_route_ports_init(p_route, IB_ENDP_MAX_NUMBER_PORTS);
	}

	memset(p_context->p_least_hops, OSM_NO_PATH, num_vertices * num_vertices);

	ret = p_context;

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return ret;
}

void osm_routing_clear_hops(IN osm_route_context_t *p_context, osm_vertex_t *p_vertex)
{
	// TODO - fill in
}

#define	bitmap_index(P_CONTEXT,P_SRC_VERTEX,P_DST_VERTEX)			\
	((P_SRC_VERTEX->index * ROUNDUP(P_CONTEXT->num_vertices,64)) + P_DST_VERTEX->index)

void osm_routing_set_least_dir(osm_route_context_t *p_context, osm_vertex_t *p_src_vertex, osm_vertex_t *p_dst_vertex, uint8_t dir)
{
	unsigned		bitno = bitmap_index(p_context, p_src_vertex, p_dst_vertex);

	if (dir == ROUTING_UP) {
		cl_bitmap_set(&p_context->least_dirs, bitno);
	} else {
		cl_bitmap_clear(&p_context->least_dirs, bitno);
	}
}

uint8_t osm_routing_get_least_dir(IN osm_route_context_t *p_context,
				  IN osm_vertex_t *p_src_vertex,
				  IN osm_vertex_t *p_dst_vertex)
{
	uint8_t			value = 0;
	unsigned		bitno = bitmap_index(p_context, p_src_vertex, p_dst_vertex);

	value = cl_bitmap_get(&p_context->least_dirs, bitno);
	return (value == 0) ? ROUTING_DOWN : ROUTING_UP;
}

uint8_t osm_routing_get_hops(IN osm_route_context_t *p_context,
			     IN osm_vertex_t *p_src_vertex,
			     IN osm_vertex_t *p_dst_vertex,
			     IN osm_edge_t *p_edge)
{
	uint8_t			dir_src_next;
	uint8_t			dir_next_dst;
	uint8_t			hops_src_dst;
	uint8_t			hops_src_next;
	uint8_t			hops_next_dst;
	osm_vertex_t		*p_next_vertex;

	/*
	 * If there isn't a path from the src -> dst, just return no path.  This check is
	 * edge independent.
	 */
	hops_src_dst = osm_routing_get_least_hops(p_context, p_src_vertex, p_dst_vertex);
	if (hops_src_dst == OSM_NO_PATH) {
		return OSM_NO_PATH;
	}

	/*
	 * If the next is the dst, just return the hop count.
	 */
	p_next_vertex = osm_graph_index_to_vertex(p_context->p_graph, p_edge->remote_vertex_index);
	if (p_next_vertex == p_dst_vertex) {
		return hops_src_dst;
	}

	hops_src_next = osm_routing_get_least_hops(p_context, p_src_vertex, p_next_vertex);
	if (hops_src_next == OSM_NO_PATH) {
		return OSM_NO_PATH;
	}

	/*
	 * If there isn't a path from the next -> dst, just return no path.
	 */
	hops_next_dst = osm_routing_get_least_hops(p_context, p_next_vertex, p_dst_vertex);
	if (hops_next_dst == OSM_NO_PATH) {
		return OSM_NO_PATH;
	}

	/*
	 * If there is no directional check, return the hop count.
	 */
	if (!p_context->p_re->dir_enabled) {
		return hops_src_next + hops_next_dst;
	}

	/*
	 * If the direction from src -> next is UP, then return the hop count.  (This is true
	 * because we know that there is a path from next -> dst and an UP or DOWN can follow
	 * an UP path.)
	 */
	dir_src_next = osm_routing_get_least_dir(p_context, p_src_vertex, p_next_vertex);
	if (dir_src_next == ROUTING_UP) {
		return hops_src_next + hops_next_dst;
	}

	/*
	 * At this point the direction from src -> next is DOWN.  So we can only have a path
	 * through next if the direction from next -> dst is DOWN.
	 */
	dir_next_dst = osm_routing_get_least_dir(p_context, p_next_vertex, p_dst_vertex);

	return (dir_next_dst == ROUTING_DOWN) ? hops_src_next + hops_next_dst : OSM_NO_PATH;
}


void osm_routing_set_least_hops(osm_route_context_t *p_context, osm_vertex_t *p_src_vertex, osm_vertex_t *p_dst_vertex, uint8_t hops)
{
	int			src_index = p_src_vertex->index;
	int			dst_index = p_dst_vertex->index;
	size_t			index = (src_index * p_context->num_vertices) + dst_index;

	p_context->p_least_hops[index] = hops;
}

uint8_t osm_routing_get_least_hops(IN osm_route_context_t *p_context,
				   IN osm_vertex_t *p_src_vertex,
				   IN osm_vertex_t *p_dst_vertex)
{
	int			src_index = p_src_vertex->index;
	int			dst_index = p_dst_vertex->index;
	int			index = (src_index * p_context->num_vertices) + dst_index;

	return p_context->p_least_hops[index];
}

uint8_t osm_routing_get_minhop_port_count(IN osm_route_context_t *p_context,
					  IN osm_vertex_t *p_src_vertex,
					  IN osm_vertex_t *p_dst_vertex)
{
	return routing_get_minhop_info(p_context, p_src_vertex, p_dst_vertex, NULL, MINHOP_PORT_COUNT);
}

uint8_t osm_routing_get_minhop_port_nums(IN osm_route_context_t *p_context,
					 IN osm_vertex_t *p_src_vertex,
					 IN osm_vertex_t *p_dst_vertex,
					 IN uint8_t *ports)
{
	return routing_get_minhop_info(p_context, p_src_vertex, p_dst_vertex, (void *)ports, MINHOP_PORT_NUMS);
}

uint8_t osm_routing_get_minhop_ports(IN osm_route_context_t *p_context,
				     IN osm_vertex_t *p_src_vertex,
				     IN osm_vertex_t *p_dst_vertex,
				     IN uint64_t *ports)
{
	return routing_get_minhop_info(p_context, p_src_vertex, p_dst_vertex, (void *)ports, MINHOP_PORT_MASK);
}
