/*
 * Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Implementation of Minhop Algorithm.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <limits.h>
#include <ctype.h>
#include <complib/cl_debug.h>
#include <complib/cl_qmap.h>
#include <complib/cl_list.h>
#include <complib/cl_heap.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTING_V2_MINHOP_C
#include <opensm/osm_switch.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_mcast_mgr.h>
#include <opensm/osm_root_detection.h>
#include <opensm/osm_max_flow_algorithm.h>
#include <opensm/osm_routing_v2.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_helper.h>

extern int osm_routing_v2_ar_dijkstra_route(IN osm_route_context_t *p_context);
extern int osm_routing_v2_ar_bfs_route(IN osm_route_context_t *p_context);

/* ----------------------------------------------------------------------------------------------
 * Hop count section.
 */

typedef struct {
	osm_ucast_mgr_t		*p_mgr;
	osm_graph_t		*p_graph;
	void			*p_data;
} hopcount_context_t;

typedef struct {
	void		*p_void;
	size_t		index;
} heap_context_t;

#define	graph_key_gen(P_VERTEX, DISTANCE)	(((uint64_t)(DISTANCE) << 32) | (P_VERTEX)->index)

static void apply_update(const void * context, const size_t new_index)
{
	heap_context_t		*p_heap_context = (heap_context_t *)context;

	if (p_heap_context) {
		p_heap_context->index = new_index;
	}
}

static int key_cmp(const void *p_lhs, const void *p_rhs)
{
	uint64_t		key1 = *(uint64_t *)p_lhs;
	uint64_t		key2 = *(uint64_t *)p_rhs;

	return (key1 < key2) ? -1 : 1;
}

static void minhop_set_hopcounts_dijkstra(IN osm_route_context_t *p_context,
					  IN osm_vertex_t *p_base_vertex,
					  IN osm_vdh_t *p_vdh)
{
	int			base;
	uint8_t			*visited;
	uint8_t			*distance;
	uint16_t		*heap_context;
	uint32_t		src_index;
	uint32_t		dst_index;
	uint32_t		src_weight;
	uint32_t		dst_weight;
	osm_graph_t		*p_graph;
	osm_vertex_t		*p_src_vertex;
	osm_vertex_t		*p_dst_vertex;
	osm_edge_t		*p_edge;
	cl_heap_t		heap;
	cl_heap_t		*p_heap;
	heap_context_t		*p_vector_context;

	p_graph = p_context->p_graph;

	osm_routing_vdh_reset(p_vdh);

	visited = p_vdh->visited;
	distance = &p_vdh->distance[p_base_vertex->index * p_graph->num_vertices];
	heap_context = p_vdh->heap;

	base = p_base_vertex->index;

	visited[base] = TRUE;
	distance[base] = 0;
	p_vector_context = (heap_context_t *)&heap_context[base];
	p_vector_context->p_void = p_base_vertex;

	/*
	 * Create the heap.
	 */
	p_heap = &heap;
	cl_heap_construct(p_heap);
	cl_heap_init(p_heap, p_graph->num_vertices + 1, 4, &apply_update, &key_cmp);
	cl_heap_insert(p_heap, graph_key_gen(p_base_vertex, 0), p_vector_context);

	/*
	 * Process the heap.
	 */
	while (!cl_heap_is_empty(p_heap)) {
		p_vector_context = (heap_context_t *)cl_heap_extract_root(p_heap);
		p_src_vertex = p_vector_context->p_void;
		src_index = p_src_vertex->index;
		src_weight = distance[src_index];

		foreach_graph_vertex_edge(p_src_vertex, p_edge) {
			dst_index = p_edge->remote_vertex_index;
			dst_weight = src_weight + p_edge->weight;
			p_dst_vertex = osm_graph_index_to_vertex(p_graph, dst_index);
			p_vector_context = (heap_context_t *)&heap_context[dst_index];

			if (!visited[dst_index]) {
				visited[dst_index] = TRUE;

				distance[p_dst_vertex->index] = dst_weight;
				p_vector_context->p_void = p_dst_vertex;
				cl_heap_insert(p_heap, graph_key_gen(p_dst_vertex, dst_weight), p_vector_context);
			} else if (dst_weight < distance[p_dst_vertex->index]) {
				distance[p_dst_vertex->index] = dst_weight;
				cl_heap_modify_key(p_heap, graph_key_gen(p_dst_vertex, dst_weight), p_vector_context->index);
			}
		}
	}

	cl_heap_destroy(p_heap);
	return;
}

/* ---------------------------------------------------------------------------------------------- */

static void minhop_set_hopcounts_bfs(IN osm_route_context_t *p_context,
				     IN osm_vertex_t *p_src_vertex,
				     IN osm_vdq_t *p_vdq)
{
	int			i;
	int			v;
	int			q_max;
	int			hop_count;
	int			remote_index;
	uint8_t			*visited;
	uint8_t			*distance;
	uint16_t		*queue;
	osm_edge_t		*p_edge;
	osm_vertex_t		*p_vertex;
	osm_vertex_t		*p_dst_vertex;
	osm_graph_t		*p_graph;

	p_graph = p_context->p_graph;

	osm_routing_vdq_reset(p_vdq);

	visited = p_vdq->visited;
	distance = &p_vdq->distance[p_src_vertex->index * p_graph->num_vertices];
	queue = p_vdq->queue;

	distance[p_src_vertex->index] = 0;
	visited[p_src_vertex->index] = 1;

	q_max = 0;
	queue[q_max++] = p_src_vertex->index;

	for (i = 0; i < q_max; i++) {
		v = queue[i];
		p_vertex = osm_graph_index_to_vertex(p_graph, v);
		hop_count = distance[v] + 1;

		foreach_graph_vertex_edge(p_vertex, p_edge) {
			remote_index = p_edge->remote_vertex_index;
			p_dst_vertex = osm_graph_index_to_vertex(p_graph, remote_index);

			if (visited[remote_index] == 0) {
				visited[remote_index] = 1;
				distance[remote_index] = hop_count;
				queue[q_max++] = remote_index;
				osm_routing_set_least_hops(p_context, p_vertex, p_dst_vertex, hop_count);
			}
		}
	}
}

/* ---------------------------------------------------------------------------------------------- */

static int minhop_set_hopcounts_parallel(IN void **tasks,
					 IN unsigned task_len,
					 IN void *context,
					 IN unsigned int thread_id)
{
	int			i;
	int			status;
	void			*p_void;
	osm_graph_t		*p_graph;
	osm_vertex_t		*p_vertex;
	osm_ucast_mgr_t		*p_mgr;
	osm_route_context_t	*p_context;

	status = 0;
	p_context = (osm_route_context_t *)context;
	p_mgr = p_context->p_mgr;
	p_graph = p_context->p_graph;
	OSM_LOG_ENTER(&p_mgr->p_subn->p_osm->log);

	if (p_graph->max_weight > 1) {
		p_void = (void *)osm_routing_vdh_init(p_mgr, p_graph);
	} else {
		p_void = (void *)osm_routing_vdq_init(p_mgr, p_graph);
	}

	if (p_void == NULL) {
		OSM_LOG(&p_mgr->p_subn->p_osm->log, OSM_LOG_ERROR, "ERR BF03: allocation failed for setting hop count data\n");
		return -1;
	}

	for (i = 0; i < task_len; i++) {
		p_vertex = (osm_vertex_t *)tasks[i];

		if (p_graph->max_weight > 1) {
			minhop_set_hopcounts_dijkstra(p_context, p_vertex, (osm_vdh_t *)p_void);
		} else {
			minhop_set_hopcounts_bfs(p_context, p_vertex, (osm_vdq_t *)p_void);
		}
	}

	if (p_graph->max_weight > 1) {
		osm_routing_vdh_free(p_void);
	} else {
		osm_routing_vdq_free(p_void);
	}

	OSM_LOG_EXIT(&p_mgr->p_subn->p_osm->log);
	return status;
}

/* ---------------------------------------------------------------------------------------------- */

__attribute__((unused))
static int minhop_set_hopcounts(IN osm_route_context_t *p_context)
{
	osm_route_context_t	context;

	/*
	 * Execute the function.
	 */
	context = *p_context;

	return osm_routing_execute_threaded(p_context, minhop_set_hopcounts_parallel, &context);
}

__attribute__((unused))
static void print_hop_counts(osm_graph_t *p_graph) {
	uint8_t			hop_count;
	uint16_t		remote_lid;
	osm_switch_t		*p_sw;
	osm_switch_t		*p_remote_sw;
	osm_vertex_t		*p_vertex;
	osm_vertex_t		*p_remote_vertex;

	foreach_graph_vertex(p_graph, p_vertex) {
		p_sw = p_vertex->p_sw;

		printf("%-6s : ", osm_switch_name(p_sw));
		foreach_graph_vertex(p_graph, p_remote_vertex) {
			p_remote_sw = p_remote_vertex->p_sw;
			remote_lid = osm_node_get_base_lid_ho(p_remote_sw->p_node, 0);
			hop_count = osm_switch_get_least_hops(p_sw, remote_lid);
			printf("%2d ", hop_count);
		}
		printf("\n");
	}
}

/* -------------------------------------------------------------------------------------------------
 * Interface functions.
 */

static int ar_minhop_route(IN void *context)
{
	int			status = -1;
	osm_route_context_t	*p_context = (osm_route_context_t *)context;
	osm_graph_t		*p_graph = p_context->p_graph;
	osm_vertex_t		*p_vertex = NULL;
	osm_switch_t		*p_sw= NULL;

	foreach_graph_vertex(p_graph, p_vertex) {
		if (p_vertex->num_hcas > 0) {
			p_sw = osm_graph_vertex_to_switch(p_vertex);
			osm_subn_add_leaf(p_context->p_mgr->p_subn, p_sw);
		}
	}

	if (p_graph->max_weight > 1) {
		status = osm_routing_v2_ar_dijkstra_route(p_context);
	} else {
		status = osm_routing_v2_ar_bfs_route(p_context);
	}

	return status;
}

static int ar_minhop_get_required_plfts(IN void *context)
{
	return 1;
}

static int ar_minhop_get_required_subgroups(IN void * context)
{
	return 1;
}

static int ar_minhop_plane_preprocess(IN void *context, IN osm_graph_t *p_graph)
{
	osm_route_context_t	*p_context = (osm_route_context_t *)context;

	minhop_set_hopcounts(p_context);
	return 0;
}

static int ar_minhop_global_preprocess(IN void *context)
{
	return 0;
}

static int ar_minhop_get_ar_mask(IN void *context,
			       IN osm_switch_t *p_src_sw,
			       IN osm_switch_t *p_dst_sw,
			       IN uint8_t plane,
			       IN int plft,
			       IN int subgroup,
			       IN uint64_t *ports)
{
	return osm_routing_get_ports(context, p_src_sw, p_dst_sw, ports);
}

static int ar_minhop_plane_postprocess(IN void *context, IN osm_graph_t *p_graph)
{
        return 0;
}

static int ar_minhop_global_postprocess(IN void *context)
{
        return 0;
}

/* -------------------------------------------------------------------------------------------------
 * API functions.
 */

int osm_routing_v2_ar_minhop_route(IN void *context)
{
	return ar_minhop_route(context);
}

int osm_routing_v2_ar_minhop_setup(IN osm_opensm_t *p_osm, IN osm_routing_engine_t *p_re)
{
	p_re->is_v2 = TRUE;
	p_re->dir_enabled = FALSE;
	p_re->route = ar_minhop_route;
	p_re->get_required_plfts = ar_minhop_get_required_plfts;
	p_re->get_required_subgroups = ar_minhop_get_required_subgroups;
	p_re->get_ar_mask = ar_minhop_get_ar_mask;
	p_re->plane_preprocess = ar_minhop_plane_preprocess;
	p_re->global_preprocess = ar_minhop_global_preprocess;
	p_re->plane_postprocess = ar_minhop_plane_postprocess;
	p_re->global_postprocess = ar_minhop_global_postprocess;

	return 0;
}

