/*
 * Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Implementation of Up Down Algorithm using ranking & Min Hop
 *      Calculation functions
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <ctype.h>
#include <complib/cl_debug.h>
#include <complib/cl_qmap.h>
#include <complib/cl_heap.h>
#include <complib/cl_ptr_vector.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTING_V2_UPDN_C
#include <opensm/osm_switch.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_mcast_mgr.h>
#include <opensm/osm_root_detection.h>
#include <opensm/osm_inform.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_routing_v2.h>

int osm_routing_v2_asym1(osm_route_context_t *p_context);
int osm_routing_v2_asym2(osm_route_context_t *p_context);
int osm_routing_v2_asym3(osm_route_context_t *p_context);

int osm_apply_func_on_port_group_nodes(char *, int pfn_func(void *, uint64_t , char *), void *, char *, void *, boolean_t);

/* ---------------------------------------------------------------------------------------------- */

static boolean_t inline updn_direction(IN osm_vertex_t *p_src_vertex, IN osm_vertex_t *p_dst_vertex)
{
	int			dir;
	osm_switch_t		*p_src_sw;
	osm_switch_t		*p_dst_sw;

	p_src_sw = osm_graph_vertex_to_switch(p_src_vertex);
	p_dst_sw = osm_graph_vertex_to_switch(p_dst_vertex);

	if (p_dst_sw->rank == p_src_sw->rank - 1) {
		dir = ROUTING_UP;
	} else if (p_dst_sw->rank == p_src_sw->rank + 1) {
		dir = ROUTING_DOWN;
	} else {
		dir = ROUTING_INVALID;
	}

	return dir;
}

static boolean_t inline updn_check_direction(IN int src_dir, IN int dir)
{
	if ((src_dir == ROUTING_INVALID) || (dir == ROUTING_INVALID)) {
		return FALSE;
	}

	if ((src_dir == ROUTING_DOWN) && (dir != ROUTING_DOWN)) {
		return FALSE;
	}

	return TRUE;
}

/* ---------------------------------------------------------------------------------------------- */

static int rank_switch_node(IN void *context, IN uint64_t guid, IN char *p)
{
        int			rank = (strlen(p) > 0) ? atoi(p) : 0;
	osm_ucast_mgr_t		*p_mgr = (osm_ucast_mgr_t *)context;
	osm_subn_t		*p_subn = p_mgr->p_subn;
	osm_switch_t		*p_sw;

	p_sw = osm_get_switch_by_guid(p_subn, cl_hton64(guid));
	if (p_sw != NULL) {
		p_sw->rank = rank;
		if (rank == 0) {
			p_mgr->num_roots++;
		}
	}

	return 0;
}

static void updn_find_root_nodes_by_min_hop_v2(IN osm_ucast_mgr_t *p_mgr, IN osm_graph_t *p_graph)
{
	int			i;
	osm_vertex_t		*p_vertex;
	osm_vertex_t		*p_remote_vertex;
	osm_switch_t		*p_sw;
	osm_switch_t		*p_remote_sw;
	osm_subn_t		*p_subn = p_mgr->sm->p_subn;
	osm_opensm_t		*p_osm = p_subn->p_osm;
	osm_log_t		*p_log = &p_osm->log;
	double			threshold_1;
	double			threshold_2;
	uint16_t		lid_ho;
	int			num_hcas;
	int			hop_count;
	int			hop_histogram[IB_SUBNET_PATH_HOPS_MAX];
	int			threshold_1_count;
	int			threshold_2_count;

	OSM_LOG_ENTER(p_log);

#define	ROOT_THRESHOLD_1	0.9
#define	ROOT_THRESHOLD_2	0.05

	num_hcas = 0;
	foreach_graph_vertex(p_graph, p_vertex) {
		num_hcas += p_vertex->num_hcas;
	}

	threshold_1 = num_hcas * ROOT_THRESHOLD_1;
	threshold_2 = num_hcas * ROOT_THRESHOLD_2;

	OSM_LOG(p_log, OSM_LOG_DEBUG, "Found %d CAs.  Thresholds are %f and %f\n",
		num_hcas, threshold_1, threshold_2);

	OSM_LOG(p_log, OSM_LOG_VERBOSE, "Passing through all vertices to collect min hop counts\n");
	foreach_graph_vertex(p_graph, p_vertex) {
		p_sw = p_vertex->p_sw;

		memset(hop_histogram, 0, sizeof(hop_histogram));
		foreach_graph_vertex(p_graph, p_remote_vertex) {
			p_remote_sw = p_remote_vertex->p_sw;
			lid_ho = osm_node_get_base_lid_ho(p_remote_sw->p_node, 0);
			hop_count = osm_switch_get_least_hops(p_sw, lid_ho);
			if (hop_count < IB_SUBNET_PATH_HOPS_MAX) {
				hop_histogram[hop_count] += p_remote_vertex->num_hcas;
			}
		}

		threshold_1_count = 0;
		threshold_2_count = 0;
		for (i = 0; i < IB_SUBNET_PATH_HOPS_MAX; i++) {
			if (hop_histogram[i] > threshold_1) {
				threshold_1_count++;
			}
			if (hop_histogram[i] > threshold_2) {
				threshold_2_count++;
			}
		}

		if ((threshold_1_count == 1) && (threshold_2_count == 1)) {
			OSM_LOG(p_log, OSM_LOG_DEBUG, "Ranking %s (0x%016lx) as root node\n",
				osm_switch_name(p_sw), osm_node_get_node_guid_ho(p_sw->p_node));
			p_sw->rank = 0;
			p_mgr->num_roots++;
		}
	}

	OSM_LOG_EXIT(p_log);
	return;
}

static int updn_set_roots(IN osm_ucast_mgr_t *p_mgr, IN osm_graph_t *p_graph)
{
	int			status = -1;
	osm_subn_t		*p_subn = p_mgr->p_subn;
	osm_opensm_t		*p_osm = p_mgr->p_subn->p_osm;

	/*
	 * If there is a root guid file, try it first.
	 */
	if (p_subn->opt.root_guid_file) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Fetching root nodes from file '%s'\n", p_subn->opt.root_guid_file);

		status = parse_node_map_str(p_subn->opt.root_guid_file, osm_apply_func_on_port_group_nodes, rank_switch_node, p_mgr, p_osm, TRUE);
		if (status != 0) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR C101: cannot parse root guids file '%s'\n", p_subn->opt.root_guid_file);
		} else if (p_mgr->num_roots == 0) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "No roots were found from parsing root guids file '%s'\n", p_subn->opt.root_guid_file);
		}
	}

	/*
	 * If there was a root guid file which we couldn't parse or no roots were found, try alternate methods.
	 */
	if ((status != 0) || (p_mgr->num_roots == 0)) {
		if (p_subn->opt.find_roots_color_algorithm) {
			osm_root_detection_find_root_nodes_by_coloring_v2(
				p_subn->p_osm,
				p_subn,
				p_graph,
				OSM_TOPOLOGY_TYPE_TREE,
				&p_mgr->num_roots, NULL);
		} else {
			updn_find_root_nodes_by_min_hop_v2(p_mgr, p_graph);
		}
	}

	return p_mgr->num_roots;
}

/* ---------------------------------------------------------------------------------------------- */

static void updn_set_ranks(IN osm_ucast_mgr_t *p_mgr, IN osm_graph_t *p_graph)
{
	int			rank;
	boolean_t		found;
	osm_switch_t		*p_sw;
	osm_switch_t		*p_remote_sw;
	osm_vertex_t		*p_vertex;
	osm_edge_t		*p_edge;

	/*
	 * Set ranks.  Cascade down the tree starting from the root nodes.   All nodes attached to
	 * the root nodes will be rank 1.  Nodes attached to rank 1 will be rank 2, etc.
	 */
	rank = -1;
	found = TRUE;
	while (found) {
		rank++;
		found = FALSE;

		foreach_graph_vertex(p_graph, p_vertex) {
			p_sw = p_vertex->p_sw;

			if (p_sw->rank == rank) {
				foreach_graph_vertex_edge(p_vertex, p_edge) {
					p_remote_sw = p_graph->vertices[p_edge->remote_vertex_index].p_sw;
					if (p_remote_sw->rank == OSM_SW_NO_RANK) {
						p_remote_sw->rank = rank+1;
						found = TRUE;
					}
				}
			}
		}
	}
}

/* ---------------------------------------------------------------------------------------------- */

static void updn_set_leaves(IN osm_ucast_mgr_t *p_mgr, IN osm_graph_t *p_graph)
{
	int			index;
	boolean_t		is_max_rank;
	osm_log_t		*p_log;
	osm_subn_t		*p_subn;
	osm_edge_t		*p_edge;
	osm_vertex_t		*p_vertex;
	osm_vertex_t		*p_remote_vertex;
	osm_switch_t		*p_sw;
	osm_switch_t		*p_remote_sw;
	cl_ptr_vector_t		candidates;
	cl_map_t		neighbors;
	cl_map_iterator_t	iter;

	p_log = p_mgr->p_log;
	p_subn = p_mgr->p_subn;

	OSM_LOG_ENTER(p_log);
	cl_map_init(&neighbors, IB_ENDP_MAX_NUMBER_PORTS);
	cl_ptr_vector_init(&candidates, 0, 64);

	/*
	 * If we aren't going to adjust ranks, then the locally maximal ranked nodes
	 * are flagged as leaves.
	 */
	foreach_graph_vertex(p_graph, p_vertex) {
		p_sw = osm_graph_vertex_to_switch(p_vertex);
		is_max_rank = TRUE;

		foreach_graph_vertex_edge(p_vertex, p_edge) {
			p_remote_sw = osm_graph_index_to_switch(p_graph, p_edge->remote_vertex_index);
			if (p_remote_sw->rank > p_sw->rank) {
				is_max_rank = FALSE;
				break;
			}
		}

		if (is_max_rank) {
			if ((p_subn->opt.routing_flags & OSM_ROUTING_FLAGS_RANK_ADJUSTMENT) == 0) {
				osm_subn_add_leaf(p_subn, p_sw);
			} else {
				cl_ptr_vector_insert(&candidates, p_vertex, NULL);
			}
		}
	}

	/*
	 * We are going to adjust ranks if necessary.  
	 */
	for (index = 0; index < cl_ptr_vector_get_size(&candidates); index++) {
		p_vertex = (osm_vertex_t *)cl_ptr_vector_get(&candidates, index);
		p_sw = osm_graph_vertex_to_switch(p_vertex);

		/*
		 * If this switch has hcas and its rank is 2 or less, then it is a leaf.
		 */
		if ((p_sw->rank < 3) || (p_vertex->num_hcas > 0)) {
			osm_subn_add_leaf(p_subn, p_sw);
			continue;
		}

		/*
		 * Count the number of neighbors which have attached HCAs.
		 */
		cl_map_remove_all(&neighbors);
		foreach_graph_vertex_edge(p_vertex, p_edge) {
			p_remote_vertex = osm_graph_index_to_vertex(p_graph, p_edge->remote_vertex_index);
			if ((cl_map_get(&neighbors, p_remote_vertex->index) == NULL) &&
			    (p_remote_vertex->num_hcas > 0)) {
				cl_map_insert(&neighbors, p_remote_vertex->index, p_remote_vertex);
			}
		}

		/*
		 * If 2 or more neighbors have HCAs, then we promote this switch up 2 ranks and 
		 * set the neighbors as leaves.
		 */
		if (cl_map_count(&neighbors) >= 2) {
			p_sw->rank -= 2;

			iter = cl_map_head(&neighbors);
			while (iter != cl_map_end(&neighbors)) {
				p_remote_vertex = (osm_vertex_t *)cl_map_obj(iter);
				p_remote_sw = osm_graph_vertex_to_switch(p_remote_vertex);
				osm_subn_add_leaf(p_subn, p_remote_sw);
				iter = cl_map_next(iter);
			}
		}
	}

	cl_map_remove_all(&neighbors);
	cl_map_destroy(&neighbors);

	cl_ptr_vector_remove_all(&candidates);
	cl_ptr_vector_destroy(&candidates);
}

/* ---------------------------------------------------------------------------------------------- */

static void updn_set_coord(IN osm_graph_t *p_graph,
			   IN osm_vertex_t *p_vertex,
			   IN uint16_t id,
			   IN osm_vertex_t **p_queue)
{
	int			i;
	int			q_max;
	uint16_t		rank;
	uint16_t		leaf_rank;
	osm_switch_t		*p_sw;
	osm_switch_t		*p_remote_sw;
	osm_vertex_t		*p_remote_vertex;
	osm_edge_t		*p_edge;

	memset(p_queue, 0, (p_graph->num_vertices + 1) * sizeof(osm_vertex_t *));

	p_sw = p_vertex->p_sw;
	p_sw->coord = id;
	leaf_rank = p_sw->rank;

	/*
	 * q_max is the index of the next entry to fill.
	 */
	q_max = 0;
	p_queue[q_max++] = p_vertex;

	/*
	 * Loop over the queue of known nodes.  When i == q_max, the loop ends because
	 * there isn't another node to process.
	 */
	for (i = 0; i < q_max; i++) {
		p_vertex = p_queue[i];
		p_sw = p_vertex->p_sw;
		rank = (p_sw->rank == leaf_rank) ? leaf_rank - 1 : leaf_rank;

		foreach_graph_vertex_edge(p_vertex, p_edge) {
			p_remote_vertex = osm_graph_index_to_vertex(p_graph, p_edge->remote_vertex_index);
			p_remote_sw = osm_graph_index_to_switch(p_graph, p_edge->remote_vertex_index);

			if ((p_remote_sw->coord == OSM_SW_NO_NEIGHBORHOOD_ID) && (p_remote_sw->rank == rank)) {
				p_remote_sw->coord = id;
				p_queue[q_max++] = p_remote_vertex;
			}
		}
	}
}

static void updn_set_coords(IN osm_ucast_mgr_t *p_mgr, IN osm_graph_t *p_graph)
{
	uint8_t			rank;
	uint8_t			max_rank;
	uint16_t		id;
	osm_switch_t		*p_sw;
	osm_vertex_t		*p_vertex;
	osm_vertex_t		**p_queue;
	cl_list_iterator_t	iter;

	OSM_LOG_ENTER(p_mgr->p_log);

	if ((p_queue = malloc((p_graph->num_vertices + 1) * sizeof(osm_vertex_t *))) == NULL) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR C102: Failed to allocate memory for queue\n");
		OSM_LOG_EXIT(p_mgr->p_log);
		return;
	}

	/* Clear the coord and use it for neighborhood id. */

	foreach_graph_vertex(p_graph, p_vertex) {
		p_sw = p_vertex->p_sw;
		p_sw->coord = OSM_SW_NO_NEIGHBORHOOD_ID;
	}

	/* Find the leaf node with the largest rank. */

	max_rank = 0;
	foreach_leaf_iter(p_mgr->p_subn, iter) {
		p_sw = leaf_iter_to_switch(iter);
		if (p_sw->rank > max_rank) {
			max_rank = p_sw->rank;
		}
	}

	/* BFS and label the neighborhoods. */

	id = 1;
	for (rank = max_rank; rank > 0; rank--) {
		foreach_leaf_iter(p_mgr->p_subn, iter) {
			p_sw = leaf_iter_to_switch(iter);

			if ((p_sw->rank == rank) && (p_sw->coord == OSM_SW_NO_NEIGHBORHOOD_ID)) {
				p_vertex = osm_graph_switch_to_vertex(p_graph, p_sw);
				updn_set_coord(p_graph, p_vertex, id, p_queue);
				id++;
			}
		}
	}

	free(p_queue);

	OSM_LOG_EXIT(p_mgr->p_log);
}

/* ---------------------------------------------------------------------------------------------- */

static int updn_set_hopcounts_dijkstra(IN osm_route_context_t *p_context,
				       IN osm_vertex_t *p_src_vertex,
				       IN osm_vddfq_t *p_vddfq)
{
	// TODO - fill in
	return 0;
}

static int updn_set_hopcounts_bfs(IN osm_route_context_t *p_context,
				  IN osm_vertex_t *p_src_vertex,
				  IN osm_vddfq_t *p_vddfq)
{
	int			i;
	int			v;
	int			dst;
	int			dir;
	int			src;
	int			src_dir;
	int			hop_count;
	int			q_max;
	uint8_t			*first;
	uint8_t			*visited;
	uint8_t			*distance;
	uint8_t			*direction;
	uint16_t		*queue;
	osm_graph_t		*p_graph;
	osm_vertex_t		*p_vertex;
	osm_vertex_t		*p_dst_vertex;
	osm_edge_t		*p_edge;

	osm_routing_set_least_hops(p_context, p_src_vertex, p_src_vertex, 0);

	visited = p_vddfq->visited;
	distance = p_vddfq->distance;
	direction = p_vddfq->direction;
	first = p_vddfq->first;
	queue = p_vddfq->queue;

	src = p_src_vertex->index;

	p_graph = p_context->p_graph;
	src_dir = ROUTING_UP;

	q_max = 0;
	queue[q_max++] = src;
	visited[src] = TRUE;
	distance[src] = 0;
	direction[src] = src_dir;;
	first[src] = 0;;

	/*
	 * BFS with direction.
	 */
	for (i = 0; i < q_max; i++) {
		v = queue[i];
		src_dir = direction[v];
		hop_count = distance[v];
		p_vertex = osm_graph_index_to_vertex(p_graph, v);

		foreach_graph_vertex_edge(p_vertex, p_edge) {
			dst = p_edge->remote_vertex_index;
			p_dst_vertex = osm_graph_index_to_vertex(p_graph, dst);

			dir = updn_direction(p_vertex, p_dst_vertex);
			if (updn_check_direction(src_dir, dir)) {
				if (!visited[dst]) {
					visited[dst] = TRUE;
					distance[dst] = hop_count + p_edge->weight;
					direction[dst] = dir;
					queue[q_max++] = dst;
					first[dst] = (first[v] == 0) ? dir : first[v];

					osm_routing_set_least_dir(p_context, p_src_vertex, p_dst_vertex, first[dst]);
					osm_routing_set_least_hops(p_context, p_src_vertex, p_dst_vertex, distance[dst]);
				}
			}
		}
	}

	return 0;
}

static int updn_set_hop_count_parallel(IN void **tasks,
                                       IN unsigned task_len,
                                       IN void *context,
                                       IN unsigned int thread_id)
{
	osm_vddfq_t		*p_vddfq;
	osm_graph_t		*p_graph;
	osm_vertex_t		*p_vertex;
	osm_route_context_t	*p_context;

	p_context = context;
	p_graph = p_context->p_graph;

	p_vddfq = osm_routing_vddfq_init(p_context->p_mgr, p_context->p_graph);
	if (p_vddfq == NULL) {
		OSM_LOG(p_context->p_mgr->p_log, OSM_LOG_ERROR, "ERR C103: allocation failed for VDDFQ\n");
		return -1;
	}

	for (int i = 0; i < task_len; i++) {
		p_vertex = (osm_vertex_t *)tasks[i];
		osm_routing_vddfq_reset(p_vddfq);

		if (p_graph->max_weight > 1) {
			updn_set_hopcounts_dijkstra(p_context, p_vertex, p_vddfq);
		} else {
			updn_set_hopcounts_bfs(p_context, p_vertex, p_vddfq);
		}
	}

	osm_routing_vddfq_free(p_vddfq);
	return 0;
}

static int updn_set_hop_counts(IN osm_route_context_t *p_context)
{
	return osm_routing_execute_threaded(p_context, updn_set_hop_count_parallel, p_context);
}


/* ---------------------------------------------------------------------------------------------- */

static int updn_init(IN osm_route_context_t *p_context)
{
	int			plane = p_context->plane;
	int			num_roots;
	osm_ucast_mgr_t		*p_mgr = p_context->p_mgr;
	osm_graph_t		*p_graph = p_mgr->graphs[plane];

	num_roots = updn_set_roots(p_mgr, p_graph);
	if (num_roots == 0) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_INFO, "disabling UPDN algorithm, no root nodes were found\n");
		return -1;
	}

	updn_set_ranks(p_mgr, p_graph);
	updn_set_leaves(p_mgr, p_graph);
	updn_set_coords(p_mgr, p_graph);
	updn_set_hop_counts(p_context);

	return 0;
}

/* ---------------------------------------------------------------------------------------------- */

static int routing_ar_updn_vertex(IN osm_route_context_t *p_context, IN osm_vertex_t *p_base_vertex)
{
	int			src;
	int			dst;
	int			dir;
	int			src_dir;
	int			dist;
	int			base;
	uint8_t			*visited;
	uint8_t			*distance;
	uint8_t			*direction;
	osm_vddq_t		*p_vddq;
	osm_route_t		*p_src_route;
	osm_route_t		*p_dst_route;
	osm_vertex_t		*p_src_vertex;
	osm_vertex_t		*p_dst_vertex;
	osm_edge_t		*p_edge;
	osm_ucast_mgr_t		*p_mgr;
	osm_graph_t		*p_graph;
	cl_heap_t		heap;
	cl_heap_t		*p_heap;

	p_mgr = p_context->p_mgr;
	p_vddq = p_context->task_context;
	p_graph = osm_ucast_mgr_plane_to_graph(p_mgr, p_context->plane);
	base = p_base_vertex->index;

	osm_routing_vddq_reset(p_vddq);
	visited = p_vddq->visited;
	distance = p_vddq->distance;
	direction = p_vddq->direction;

	/*
	 * Create the heap.
	 */
	p_heap = &heap;
	cl_heap_construct(p_heap);
	cl_heap_init(p_heap, p_graph->num_vertices + 1, 4, &osm_routing_index_update, &osm_routing_key_cmp);

	/*
	 * Seed the heap with the src switch's neighbors.
	 */
	foreach_graph_vertex_edge(p_base_vertex, p_edge) {
		src_dir = ROUTING_UP;
		dist = p_edge->weight;

		dst = p_edge->remote_vertex_index;
		p_dst_vertex = osm_graph_index_to_vertex(p_graph, dst);
		p_dst_route = osm_vertex_to_vertex_route(p_context, p_base_vertex, p_dst_vertex);

		dir = updn_direction(p_base_vertex, p_dst_vertex);
		if (!updn_check_direction(src_dir, dir)) {
			continue;
		}

		if (!visited[dst]) {
			visited[dst] = TRUE;
			direction[dst] = dir;

			distance[dst] = dist;
			osm_route_ports_set(p_dst_route, p_edge->port_num);
			cl_heap_insert(p_heap, routing_key(p_dst_route, dist), p_dst_route);
		} else if (dist < distance[dst]) {
			distance[dst] = dist;

			osm_route_ports_clear(p_dst_route);
			osm_route_ports_set(p_dst_route, p_edge->port_num);

			cl_heap_modify_key(p_heap, routing_key(p_dst_route, dist), p_dst_route->heap_index);
		} else if (dist == distance[dst]) {
			osm_route_ports_set(p_dst_route, p_edge->port_num);
		}
	}

	/*
	 * Process the heap.
	 */
	while (!cl_heap_is_empty(p_heap)) {
		p_src_route = (osm_route_t *)cl_heap_extract_root(p_heap);
		p_src_route->heap_index = 0xffffffff;

		src = p_src_route->dst;
		src_dir = direction[src];
		p_src_vertex = osm_graph_index_to_vertex(p_graph, src);

		foreach_graph_vertex_edge(p_src_vertex, p_edge) {
			dist = distance[src] + p_edge->weight;

			dst = p_edge->remote_vertex_index;
			p_dst_vertex = osm_graph_index_to_vertex(p_graph, dst);
			p_dst_route = osm_vertex_to_vertex_route(p_context, p_base_vertex, p_dst_vertex);

			if (dst == base) {
				continue;
			}

			dir = updn_direction(p_src_vertex, p_dst_vertex);
			if (!updn_check_direction(src_dir, dir)) {
				continue;
			}

			if (!visited[dst]) {
				visited[dst] = TRUE;
				direction[dst] = dir;

				distance[dst] = dist;
				osm_route_ports_copy(p_dst_route, p_src_route);
				cl_heap_insert(p_heap, routing_key(p_dst_route, dist), p_dst_route);
			} else if (dist < distance[dst]) {
				distance[dst] = dist;
				osm_route_ports_copy(p_dst_route, p_src_route);

				cl_heap_modify_key(p_heap, routing_key(p_dst_route, dist), p_dst_route->heap_index);
			} else if (dist == distance[dst]) {
				osm_route_ports_merge(p_dst_route, p_src_route);
			}
		}
	}

	cl_heap_destroy(p_heap);
	return 0;
}

/* ---------------------------------------------------------------------------------------------- */

static void *updn_task_init(IN osm_route_context_t *p_context)
{
	return osm_routing_vddq_init(p_context->p_mgr, p_context->p_graph);
}

static void updn_task_fini(IN void *p_void)
{
	osm_routing_vddq_free(p_void);
}

/* ---------------------------------------------------------------------------------------------- */

static int routing_v2_ar_updn(IN osm_route_context_t *p_context)
{
	int			status;

	p_context->sw_function = routing_ar_updn_vertex;
	p_context->sw_init_function = updn_task_init;
	p_context->sw_fini_function = updn_task_fini;
	status = osm_routing_per_vertex(p_context);

	return status;
}

/* -------------------------------------------------------------------------------------------------
 * Interface functions.
 */

static int ar_updn_route(IN void *context)
{
	int			status = -1;
	osm_route_context_t	*p_context = (osm_route_context_t *)context;
	osm_ucast_mgr_t		*p_mgr = p_context->p_mgr;

	/*
	 * Initialize UPDN for all ASYM types.
	 */
	status = updn_init(p_context);
	if (status != 0) {
		return status;
	}

	/*
	 * Choose the algorithmic function.
	 */
	switch (p_mgr->p_subn->opt.ar_tree_asymmetric_flow) {
	case OSM_ASYMMETRIC_TREE_DISABLED:
		status = routing_v2_ar_updn(p_context);
		break;
	case OSM_ASYMMETRIC_TREE_1_SUB_GROUP_ENABLED:
		status = osm_routing_v2_asym1(p_context);
		break;
	case OSM_ASYMMETRIC_TREE_2_SUB_GROUPS_ENABLED:
		status = osm_routing_v2_asym2(p_context);
		break;
	case OSM_ASYMMETRIC_TREE_ASYMTREE:
		status = osm_routing_v2_asym3(p_context);
		break;
	default:
		status = routing_v2_ar_updn(p_context);
		break;
	}

	/*
	 * If asym=3 failed, fallback to asym=1.
	 */
	if (p_mgr->p_subn->opt.ar_tree_asymmetric_flow == OSM_ASYMMETRIC_TREE_ASYMTREE) {
		if (status != 0) {
			status = osm_routing_v2_asym1(p_context);
		}
	}

	return status;
}

static int ar_updn_get_required_plfts(IN void *context)
{
	osm_route_context_t	*p_context = (osm_route_context_t *)context;
	osm_ucast_mgr_t		*p_mgr = p_context->p_mgr;

	/*
	 * Choose the algorithmic function.
	 */
	switch (p_mgr->p_subn->opt.ar_tree_asymmetric_flow) {
	case OSM_ASYMMETRIC_TREE_DISABLED:
		return 1;
	case OSM_ASYMMETRIC_TREE_1_SUB_GROUP_ENABLED:
		return 1;
	case OSM_ASYMMETRIC_TREE_2_SUB_GROUPS_ENABLED:
		return 1;
	case OSM_ASYMMETRIC_TREE_ASYMTREE:
		return 1;
	default:
		return 1;
	}
}

static int ar_updn_get_required_subgroups(IN void * context)
{
	osm_route_context_t	*p_context = (osm_route_context_t *)context;
	osm_ucast_mgr_t		*p_mgr = p_context->p_mgr;

	/*
	 * Choose the algorithmic function.
	 */
	switch (p_mgr->p_subn->opt.ar_tree_asymmetric_flow) {
	case OSM_ASYMMETRIC_TREE_DISABLED:
		return 1;
	case OSM_ASYMMETRIC_TREE_1_SUB_GROUP_ENABLED:
		return 1;
	case OSM_ASYMMETRIC_TREE_2_SUB_GROUPS_ENABLED:
		return 2;
	case OSM_ASYMMETRIC_TREE_ASYMTREE:
		return 1;
	default:
		return 1;
	}
}

static int ar_updn_plane_preprocess(IN void *context, IN osm_graph_t *p_graph)
{
	return 0;
}

static int ar_updn_global_preprocess(IN void *context)
{
	return 0;
}

static int ar_updn_get_ar_mask(IN void *context,
			       IN osm_switch_t *p_src_sw,
			       IN osm_switch_t *p_dst_sw,
			       IN uint8_t plane,
			       IN int plft,
			       IN int subgroup,
			       IN uint64_t *ports)
{
	return osm_routing_get_ports(context, p_src_sw, p_dst_sw, ports);
}

static int ar_updn_plane_postprocess(IN void *context, IN osm_graph_t *p_graph)
{
        return 0;
}

static int ar_updn_global_postprocess(IN void *context)
{
        return 0;
}

/* -------------------------------------------------------------------------------------------------
 * API functions.
 */

int osm_routing_v2_ar_updn_setup(IN osm_opensm_t *p_osm, IN osm_routing_engine_t *p_re)
{
	p_re->is_v2 = TRUE;
	p_re->dir_enabled = TRUE;
	p_re->route = ar_updn_route;
	p_re->get_required_plfts = ar_updn_get_required_plfts;
	p_re->get_required_subgroups = ar_updn_get_required_subgroups;
	p_re->get_ar_mask = ar_updn_get_ar_mask;
	p_re->plane_preprocess = ar_updn_plane_preprocess;
	p_re->global_preprocess = ar_updn_global_preprocess;
	p_re->plane_postprocess = ar_updn_plane_postprocess;
	p_re->global_postprocess = ar_updn_global_postprocess;

	return 0;
}
