/*
 * Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *      Utilities for routing engines to use.
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <math.h>
#include <regex.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/time.h>

#include <complib/cl_types_osd.h>
#include <complib/cl_types.h>
#include <complib/cl_byteswap.h>
#include <complib/cl_heap.h>
#include <complib/cl_bitmap.h>

#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_ROUTING_V2_UTILS_C
#include <opensm/osm_subnet.h>
#include <opensm/osm_opensm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_routing.h>
#include <opensm/osm_advanced_routing.h>
#include <opensm/osm_graph.h>
#include <opensm/osm_ucast_mgr.h>
#include <opensm/osm_routing_v2.h>
#include <opensm/osm_helper.h>

/* ----------------------------------------------------------------------------------------------
 * Data struct section.
 */

/* ---------------------------------------------------------------------------------------------- */

void osm_routing_vdq_reset(IN osm_vdq_t *p_vdq)
{
	memset(p_vdq->visited,            0, p_vdq->visited_size);
	memset(p_vdq->distance, OSM_NO_PATH, p_vdq->distance_size);
	memset(p_vdq->queue,              0, p_vdq->queue_size);
}

void osm_routing_vdq_free(IN osm_vdq_t *p_vdq)
{
	if (p_vdq != NULL) {
		free(p_vdq->visited);
		free(p_vdq->distance);
		free(p_vdq->queue);

		free(p_vdq);
	}
}

void *osm_routing_vdq_init(IN osm_ucast_mgr_t *p_mgr, IN osm_graph_t *p_graph)
{
	osm_vdq_t		*p_vdq;

	p_vdq = calloc(1, sizeof(osm_vdq_t));
	if (p_vdq == NULL) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR BF05: allocation of VDQ failed\n");
		return NULL;
	}

	p_vdq->visited_size  = p_graph->num_vertices * sizeof(uint8_t);
	p_vdq->distance_size = p_graph->num_vertices * sizeof(uint8_t);
	p_vdq->queue_size    = p_graph->num_vertices * sizeof(uint16_t);

	p_vdq->visited  = calloc(1, p_vdq->visited_size);
	p_vdq->distance = calloc(1, p_vdq->distance_size);
	p_vdq->queue    = calloc(1, p_vdq->queue_size);

	if ((p_vdq->visited == NULL) || (p_vdq->distance == NULL) || (p_vdq->queue == NULL)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR BF06: allocation of VDQ structs failed\n");
		free(p_vdq->visited);
		free(p_vdq->distance);
		free(p_vdq->queue);
		free(p_vdq);

		p_vdq = NULL;
	}

	return p_vdq;
}

/* ---------------------------------------------------------------------------------------------- */

void osm_routing_vddq_reset(IN osm_vddq_t *p_vddq)
{
	memset(p_vddq->visited,              0, p_vddq->visited_size);
	memset(p_vddq->distance,   OSM_NO_PATH, p_vddq->distance_size);
	memset(p_vddq->direction,            0, p_vddq->direction_size);
	memset(p_vddq->queue,                0, p_vddq->queue_size);
}

void osm_routing_vddq_free(IN osm_vddq_t *p_vddq)
{
	if (p_vddq != NULL) {
		free(p_vddq->visited);
		free(p_vddq->distance);
		free(p_vddq->direction);
		free(p_vddq->queue);

		free(p_vddq);
	}
}

void *osm_routing_vddq_init(IN osm_ucast_mgr_t *p_mgr, IN osm_graph_t *p_graph)
{
	osm_vddq_t		*p_vddq;

	p_vddq = calloc(1, sizeof(osm_vddq_t));
	if (p_vddq == NULL) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR BF07: allocation of VDDQ failed\n");
		return NULL;
	}

	p_vddq->visited_size   = p_graph->num_vertices * sizeof(uint8_t);
	p_vddq->distance_size  = p_graph->num_vertices * sizeof(uint8_t);
	p_vddq->direction_size = p_graph->num_vertices * sizeof(uint8_t);
	p_vddq->queue_size     = p_graph->num_vertices * sizeof(uint16_t);

	p_vddq->visited   = calloc(1, p_vddq->visited_size);
	p_vddq->distance  = calloc(1, p_vddq->distance_size);
	p_vddq->direction = calloc(1, p_vddq->direction_size);
	p_vddq->queue     = calloc(1, p_vddq->queue_size);

	if ((p_vddq->visited == NULL) || (p_vddq->distance == NULL) || (p_vddq->direction == NULL) || (p_vddq->queue == NULL)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR BF08: allocation of VDDQ structs failed\n");
		free(p_vddq->visited);
		free(p_vddq->distance);
		free(p_vddq->direction);
		free(p_vddq->queue);
		free(p_vddq);

		p_vddq = NULL;
	}

	return p_vddq;
}

/* ---------------------------------------------------------------------------------------------- */

void osm_routing_vddfq_reset(IN osm_vddfq_t *p_vddfq)
{
	memset(p_vddfq->visited,              0, p_vddfq->visited_size);
	memset(p_vddfq->distance,   OSM_NO_PATH, p_vddfq->distance_size);
	memset(p_vddfq->direction,            0, p_vddfq->direction_size);
	memset(p_vddfq->first,                0, p_vddfq->first_size);
	memset(p_vddfq->queue,                0, p_vddfq->queue_size);
}

void osm_routing_vddfq_free(IN osm_vddfq_t *p_vddfq)
{
	if (p_vddfq != NULL) {
		free(p_vddfq->visited);
		free(p_vddfq->distance);
		free(p_vddfq->direction);
		free(p_vddfq->first);
		free(p_vddfq->queue);

		free(p_vddfq);
	}
}

void *osm_routing_vddfq_init(IN osm_ucast_mgr_t *p_mgr, IN osm_graph_t *p_graph)
{
	osm_vddfq_t		*p_vddfq;

	p_vddfq = calloc(1, sizeof(osm_vddfq_t));
	if (p_vddfq == NULL) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR BF07: allocation of VDDQ failed\n");
		return NULL;
	}

	p_vddfq->visited_size   = p_graph->num_vertices * sizeof(uint8_t);
	p_vddfq->distance_size  = p_graph->num_vertices * sizeof(uint8_t);
	p_vddfq->direction_size = p_graph->num_vertices * sizeof(uint8_t);
	p_vddfq->first_size     = p_graph->num_vertices * sizeof(uint8_t);
	p_vddfq->queue_size     = p_graph->num_vertices * sizeof(uint16_t);

	p_vddfq->visited   = calloc(1, p_vddfq->visited_size);
	p_vddfq->distance  = calloc(1, p_vddfq->distance_size);
	p_vddfq->direction = calloc(1, p_vddfq->direction_size);
	p_vddfq->first     = calloc(1, p_vddfq->first_size);
	p_vddfq->queue     = calloc(1, p_vddfq->queue_size);

	if (osm_any_null(5, p_vddfq->visited, p_vddfq->distance, p_vddfq->direction, p_vddfq->first, p_vddfq->queue)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR BF08: allocation of VDDQ structs failed\n");
		osm_routing_vddfq_free(p_vddfq);
		p_vddfq = NULL;
	}

	return p_vddfq;
}

/* ---------------------------------------------------------------------------------------------- */

void osm_routing_vdh_reset(IN osm_vdh_t *p_vdh)
{
	memset(p_vdh->visited,            0, p_vdh->visited_size);
	memset(p_vdh->distance, OSM_NO_PATH, p_vdh->distance_size);
	memset(p_vdh->heap,               0, p_vdh->heap_size);
}

void osm_routing_vdh_free(IN osm_vdh_t *p_vdh)
{
	if (p_vdh != NULL) {
		free(p_vdh->visited);
		free(p_vdh->distance);
		free(p_vdh->heap);

		free(p_vdh);
	}
}

void *osm_routing_vdh_init(IN osm_ucast_mgr_t *p_mgr, IN osm_graph_t *p_graph)
{
	osm_vdh_t		*p_vdh;

	p_vdh = calloc(1, sizeof(osm_vdh_t));
	if (p_vdh == NULL) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR BF05: allocation of VDQ failed\n");
		return NULL;
	}

	p_vdh->visited_size  = p_graph->num_vertices * sizeof(uint8_t);
	p_vdh->distance_size = p_graph->num_vertices * sizeof(uint8_t);
	p_vdh->heap_size     = p_graph->num_vertices * (sizeof(void *) + sizeof(size_t));

	p_vdh->visited  = calloc(1, p_vdh->visited_size);
	p_vdh->distance = calloc(1, p_vdh->distance_size);
	p_vdh->heap     = calloc(1, p_vdh->heap_size);

	if ((p_vdh->visited == NULL) || (p_vdh->distance == NULL) || (p_vdh->heap == NULL)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR BF06: allocation of VDQ structs failed\n");
		free(p_vdh->visited);
		free(p_vdh->distance);
		free(p_vdh->heap);
		free(p_vdh);

		p_vdh = NULL;
	}

	return p_vdh;
}

/* ----------------------------------------------------------------------------------------------
 * Hop count section.
 */

__attribute__((unused))
static void routing_hop_count_print(IN osm_ucast_mgr_t *p_mgr, IN osm_graph_t *p_graph)
{
	// TODO - fill this in
}

/* ---------------------------------------------------------------------------------------------- */

static int routing_set_hop_count_by_vertex_edge(IN osm_ucast_mgr_t *p_mgr,
					        IN osm_graph_t *p_graph,
					        IN osm_vertex_t *p_base_vertex,
					        IN osm_edge_t *p_base_edge,
					        IN osm_vdq_t *p_vdq)
{
	int			i;
	int			v;
	int			dst;
	int			base;
	int			hop_count;
	int			q_max;
	int			neighbor;
	uint8_t			*distance;
	uint8_t			*visited;
	uint16_t		*queue;
	uint16_t		lid_ho;
	osm_vertex_t		*p_vertex;
	osm_vertex_t		*p_dst_vertex;
	osm_edge_t		*p_edge;
	osm_switch_t		*p_dst_sw;
	osm_switch_t		*p_base_sw;
	osm_switch_t		*p_neighbor_sw;

	base = p_base_vertex->index;
	p_base_sw = osm_graph_index_to_switch(p_graph, base);

	visited = p_vdq->visited;
	distance = p_vdq->distance;
	queue = p_vdq->queue;

	neighbor = p_base_edge->remote_vertex_index;
	p_neighbor_sw = osm_graph_index_to_switch(p_graph, neighbor);

	lid_ho = osm_node_get_base_lid_ho(p_neighbor_sw->p_node, 0);
	osm_switch_set_hops(p_base_sw, lid_ho, p_base_edge->port_num, 1);

	q_max = 0;
	queue[q_max++] = neighbor;
	visited[neighbor] = TRUE;
	distance[neighbor] = 1;

	/*
	 * BFS.
	 */
	for (i = 0; i < q_max; i++) {
		v = queue[i];
		hop_count = distance[v];
		p_vertex = osm_graph_index_to_vertex(p_graph, v);

		foreach_graph_vertex_edge(p_vertex, p_edge) {
			dst = p_edge->remote_vertex_index;
			p_dst_vertex = osm_graph_index_to_vertex(p_graph, dst);
			p_dst_sw = osm_graph_vertex_to_switch(p_dst_vertex);

			if (!visited[dst]) {
				visited[dst] = TRUE;
				distance[dst] = hop_count + 1;
				queue[q_max++] = dst;

				lid_ho = osm_node_get_base_lid_ho(p_dst_sw->p_node, 0);
				osm_switch_set_hops(p_base_sw, lid_ho, p_base_edge->port_num, distance[dst]);
			}
		}
	}

	return 0;
}

static int routing_set_hop_count_by_vertex(IN osm_ucast_mgr_t *p_mgr,
					   IN osm_graph_t *p_graph,
					   IN osm_vertex_t *p_base_vertex,
					   IN osm_vdq_t *p_vdq)
{
	int			base;
	uint16_t		base_lid_ho;
	osm_edge_t		*p_base_edge;
	osm_switch_t		*p_base_sw;

	/*
	 * Set hop count to self.
	 */
	base = p_base_vertex->index;
	p_base_sw = osm_graph_index_to_switch(p_graph, base);

	base_lid_ho = osm_node_get_base_lid_ho(p_base_sw->p_node, 0);
	osm_switch_set_hops(p_base_sw, base_lid_ho, 0, 0);

	/*
	 * Run a BFS for each edge.
	 */
	foreach_graph_vertex_edge(p_base_vertex, p_base_edge) {
		osm_routing_vdq_reset(p_vdq);
		routing_set_hop_count_by_vertex_edge(p_mgr, p_graph, p_base_vertex, p_base_edge, p_vdq);

	}

	return 0;
}

int osm_routing_set_default_hopcounts(IN osm_ucast_mgr_t *p_mgr, IN osm_graph_t *p_graph)
{
	osm_vertex_t		*p_vertex;
	osm_vdq_t		*p_vdq;

	p_vdq = osm_routing_vdq_init(p_mgr, p_graph);
	if (p_vdq == NULL) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR C104: allocation of VDDQ failed\n");
		return -1;
	}

	foreach_graph_vertex(p_graph, p_vertex) {
		osm_routing_vdq_reset(p_vdq);
		routing_set_hop_count_by_vertex(p_mgr, p_graph, p_vertex, p_vdq);
	}

	osm_routing_vdq_free(p_vdq);
	return 0;
}
